from __future__ import annotations as _annotations

import io
from collections.abc import AsyncGenerator, AsyncIterable, AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Literal, cast, overload

from pydantic import TypeAdapter
from typing_extensions import assert_never

from upsonic.tools.builtin_tools import CodeExecutionTool, WebSearchTool

from upsonic import _utils, usage
from upsonic._utils import guard_tool_call_id as _guard_tool_call_id
from upsonic.utils.package.exception import UserError, UnexpectedModelBehavior, ModelHTTPError
from upsonic.messages import (
    BinaryContent,
    BuiltinToolCallPart,
    BuiltinToolReturnPart,
    DocumentUrl,
    FinishReason,
    ImageUrl,
    ModelMessage,
    ModelRequest,
    ModelResponse,
    ModelResponsePart,
    ModelResponseStreamEvent,
    RetryPromptPart,
    SystemPromptPart,
    TextPart,
    ThinkingPart,
    ToolCallPart,
    ToolReturnPart,
    UserPromptPart,
)
from upsonic.profiles import ModelProfileSpec
from upsonic.providers import Provider, infer_provider
from upsonic.providers.anthropic import AsyncAnthropicClient
from upsonic.models.settings import ModelSettings
from upsonic.tools import ToolDefinition
from upsonic.models import Model, ModelRequestParameters, StreamedResponse, check_allow_model_requests, download_item, get_user_agent

_FINISH_REASON_MAP: dict[BetaStopReason, FinishReason] = {
    'end_turn': 'stop',
    'max_tokens': 'length',
    'stop_sequence': 'stop',
    'tool_use': 'tool_call',
    'pause_turn': 'stop',
    'refusal': 'content_filter',
}


try:
    from anthropic import NOT_GIVEN, APIStatusError, AsyncStream
    from anthropic.types.beta import (
        BetaBase64PDFBlockParam,
        BetaBase64PDFSourceParam,
        BetaCitationsDelta,
        BetaCodeExecutionTool20250522Param,
        BetaCodeExecutionToolResultBlock,
        BetaCodeExecutionToolResultBlockContent,
        BetaCodeExecutionToolResultBlockParam,
        BetaCodeExecutionToolResultBlockParamContentParam,
        BetaContentBlock,
        BetaContentBlockParam,
        BetaImageBlockParam,
        BetaInputJSONDelta,
        BetaMessage,
        BetaMessageParam,
        BetaMetadataParam,
        BetaPlainTextSourceParam,
        BetaRawContentBlockDeltaEvent,
        BetaRawContentBlockStartEvent,
        BetaRawContentBlockStopEvent,
        BetaRawMessageDeltaEvent,
        BetaRawMessageStartEvent,
        BetaRawMessageStopEvent,
        BetaRawMessageStreamEvent,
        BetaRedactedThinkingBlock,
        BetaRedactedThinkingBlockParam,
        BetaServerToolUseBlock,
        BetaServerToolUseBlockParam,
        BetaSignatureDelta,
        BetaStopReason,
        BetaTextBlock,
        BetaTextBlockParam,
        BetaTextDelta,
        BetaThinkingBlock,
        BetaThinkingBlockParam,
        BetaThinkingConfigParam,
        BetaThinkingDelta,
        BetaToolChoiceParam,
        BetaToolParam,
        BetaToolResultBlockParam,
        BetaToolUnionParam,
        BetaToolUseBlock,
        BetaToolUseBlockParam,
        BetaWebSearchTool20250305Param,
        BetaWebSearchToolResultBlock,
        BetaWebSearchToolResultBlockContent,
        BetaWebSearchToolResultBlockParam,
        BetaWebSearchToolResultBlockParamContentParam,
    )
    from anthropic.types.beta.beta_web_search_tool_20250305_param import UserLocation
    from anthropic.types.model_param import ModelParam
    _ANTHROPIC_AVAILABLE = True
except ImportError:
    # Optional imports for Anthropic model
    # If these are not available, the model instantiation will raise an error
    _ANTHROPIC_AVAILABLE = False
    ModelParam = None  # type: ignore
    pass

LatestAnthropicModelNames = ModelParam
"""Latest Anthropic models."""

AnthropicModelName = str | LatestAnthropicModelNames
"""Possible Anthropic model names.

Since Anthropic supports a variety of date-stamped models, we explicitly list the latest models but
allow any name in the type hints.
See [the Anthropic docs](https://docs.anthropic.com/en/docs/about-claude/models) for a full list.
"""


class AnthropicModelSettings(ModelSettings, total=False):
    """Settings used for an Anthropic model request."""

    # ALL FIELDS MUST BE `anthropic_` PREFIXED SO YOU CAN MERGE THEM WITH OTHER MODELS.

    anthropic_metadata: BetaMetadataParam
    """An object describing metadata about the request.

    Contains `user_id`, an external identifier for the user who is associated with the request.
    """

    anthropic_thinking: BetaThinkingConfigParam
    """Determine whether the model should generate a thinking block.

    See [the Anthropic docs](https://docs.anthropic.com/en/docs/build-with-claude/extended-thinking) for more information.
    """


@dataclass(init=False)
class AnthropicModel(Model):
    """A model that uses the Anthropic API.

    Internally, this uses the [Anthropic Python client](https://github.com/anthropics/anthropic-sdk-python) to interact with the API.

    Apart from `__init__`, all methods are private or match those of the base class.
    """

    client: AsyncAnthropicClient = field(repr=False)

    _model_name: AnthropicModelName = field(repr=False)
    _provider: Provider[AsyncAnthropicClient] = field(repr=False)

    def __init__(
        self,
        model_name: AnthropicModelName,
        *,
        provider: Literal['anthropic'] | Provider[AsyncAnthropicClient] = 'anthropic',
        profile: ModelProfileSpec | None = None,
        settings: ModelSettings | None = None,
    ):
        """Initialize an Anthropic model.

        Args:
            model_name: The name of the Anthropic model to use. List of model names available
                [here](https://docs.anthropic.com/en/docs/about-claude/models).
            provider: The provider to use for the Anthropic API. Can be either the string 'anthropic' or an
                instance of `Provider[AsyncAnthropicClient]`. If not provided, the other parameters will be used.
            profile: The model profile to use. Defaults to a profile picked by the provider based on the model name.
            settings: Default model settings for this model instance.
        """
        if not _ANTHROPIC_AVAILABLE:
            from upsonic.utils.printing import import_error
            import_error(
                package_name="anthropic",
                install_command='pip install anthropic',
                feature_name="Anthropic model"
            )

        self._model_name = model_name

        if isinstance(provider, str):
            provider = infer_provider(provider)
        self._provider = provider
        self.client = provider.client

        super().__init__(settings=settings, profile=profile or provider.model_profile)

    @property
    def base_url(self) -> str:
        return str(self.client.base_url)

    @property
    def model_name(self) -> AnthropicModelName:
        """The model name."""
        return self._model_name

    @property
    def system(self) -> str:
        """The model provider."""
        return self._provider.name

    async def request(
        self,
        messages: list[ModelMessage],
        model_settings: ModelSettings | None,
        model_request_parameters: ModelRequestParameters,
    ) -> ModelResponse:
        check_allow_model_requests()
        response = await self._messages_create(
            messages, False, cast(AnthropicModelSettings, model_settings or {}), model_request_parameters
        )
        model_response = self._process_response(response)
        return model_response

    @asynccontextmanager
    async def request_stream(
        self,
        messages: list[ModelMessage],
        model_settings: ModelSettings | None,
        model_request_parameters: ModelRequestParameters,
    ) -> AsyncIterator[StreamedResponse]:
        check_allow_model_requests()
        response = await self._messages_create(
            messages, True, cast(AnthropicModelSettings, model_settings or {}), model_request_parameters
        )
        async with response:
            yield await self._process_streamed_response(response, model_request_parameters)

    @overload
    async def _messages_create(
        self,
        messages: list[ModelMessage],
        stream: Literal[True],
        model_settings: AnthropicModelSettings,
        model_request_parameters: ModelRequestParameters,
    ) -> AsyncStream[BetaRawMessageStreamEvent]:
        pass

    @overload
    async def _messages_create(
        self,
        messages: list[ModelMessage],
        stream: Literal[False],
        model_settings: AnthropicModelSettings,
        model_request_parameters: ModelRequestParameters,
    ) -> BetaMessage:
        pass

    async def _messages_create(
        self,
        messages: list[ModelMessage],
        stream: bool,
        model_settings: AnthropicModelSettings,
        model_request_parameters: ModelRequestParameters,
    ) -> BetaMessage | AsyncStream[BetaRawMessageStreamEvent]:
        tools = self._get_tools(model_request_parameters)
        builtin_tools, tool_headers = self._get_builtin_tools(model_request_parameters)
        tools += builtin_tools

        tool_choice: BetaToolChoiceParam | None

        if not tools:
            tool_choice = None
        else:
            if not model_request_parameters.allow_text_output:
                tool_choice = {'type': 'any'}
            else:
                tool_choice = {'type': 'auto'}

            if (allow_parallel_tool_calls := model_settings.get('parallel_tool_calls')) is not None:
                tool_choice['disable_parallel_tool_use'] = not allow_parallel_tool_calls

        system_prompt, anthropic_messages = await self._map_message(messages)

        try:
            extra_headers = model_settings.get('extra_headers', {})
            for k, v in tool_headers.items():
                extra_headers.setdefault(k, v)
            extra_headers.setdefault('User-Agent', get_user_agent())

            return await self.client.beta.messages.create(
                max_tokens=model_settings.get('max_tokens', 4096),
                system=system_prompt or NOT_GIVEN,
                messages=anthropic_messages,
                model=self._model_name,
                tools=tools or NOT_GIVEN,
                tool_choice=tool_choice or NOT_GIVEN,
                stream=stream,
                thinking=model_settings.get('anthropic_thinking', NOT_GIVEN),
                stop_sequences=model_settings.get('stop_sequences', NOT_GIVEN),
                temperature=model_settings.get('temperature', NOT_GIVEN),
                top_p=model_settings.get('top_p', NOT_GIVEN),
                timeout=model_settings.get('timeout', NOT_GIVEN),
                metadata=model_settings.get('anthropic_metadata', NOT_GIVEN),
                extra_headers=extra_headers,
                extra_body=model_settings.get('extra_body'),
            )
        except APIStatusError as e:
            if (status_code := e.status_code) >= 400:
                raise ModelHTTPError(status_code=status_code, model_name=self.model_name, body=e.body) from e
            raise  # pragma: lax no cover

    def _process_response(self, response: BetaMessage) -> ModelResponse:
        """Process a non-streamed response, and prepare a message to return."""
        items: list[ModelResponsePart] = []
        for item in response.content:
            if isinstance(item, BetaTextBlock):
                items.append(TextPart(content=item.text))
            elif isinstance(item, BetaServerToolUseBlock):
                items.append(_map_server_tool_use_block(item, self.system))
            elif isinstance(item, BetaWebSearchToolResultBlock):
                items.append(_map_web_search_tool_result_block(item, self.system))
            elif isinstance(item, BetaCodeExecutionToolResultBlock):
                items.append(_map_code_execution_tool_result_block(item, self.system))
            elif isinstance(item, BetaRedactedThinkingBlock):
                items.append(
                    ThinkingPart(id='redacted_thinking', content='', signature=item.data, provider_name=self.system)
                )
            elif isinstance(item, BetaThinkingBlock):
                items.append(ThinkingPart(content=item.thinking, signature=item.signature, provider_name=self.system))
            else:
                assert isinstance(item, BetaToolUseBlock), f'unexpected item type {type(item)}'
                items.append(
                    ToolCallPart(
                        tool_name=item.name,
                        args=cast(dict[str, Any], item.input),
                        tool_call_id=item.id,
                    )
                )

        finish_reason: FinishReason | None = None
        provider_details: dict[str, Any] | None = None
        if raw_finish_reason := response.stop_reason:  # pragma: no branch
            provider_details = {'finish_reason': raw_finish_reason}
            finish_reason = _FINISH_REASON_MAP.get(raw_finish_reason)

        return ModelResponse(
            parts=items,
            usage=_map_usage(response),
            model_name=response.model,
            provider_response_id=response.id,
            provider_name=self._provider.name,
            finish_reason=finish_reason,
            provider_details=provider_details,
        )

    async def _process_streamed_response(
        self, response: AsyncStream[BetaRawMessageStreamEvent], model_request_parameters: ModelRequestParameters
    ) -> StreamedResponse:
        peekable_response = _utils.PeekableAsyncStream(response)
        first_chunk = await peekable_response.peek()
        if isinstance(first_chunk, _utils.Unset):
            raise UnexpectedModelBehavior('Streamed response ended without content or tool calls')  # pragma: no cover

        assert isinstance(first_chunk, BetaRawMessageStartEvent)

        return AnthropicStreamedResponse(
            model_request_parameters=model_request_parameters,
            _model_name=first_chunk.message.model,
            _response=peekable_response,
            _timestamp=_utils.now_utc(),
            _provider_name=self._provider.name,
        )

    def _get_tools(self, model_request_parameters: ModelRequestParameters) -> list[BetaToolParam]:
        return [self._map_tool_definition(r) for r in model_request_parameters.tool_defs.values()]

    def _get_builtin_tools(
        self, model_request_parameters: ModelRequestParameters
    ) -> tuple[list[BetaToolUnionParam], dict[str, str]]:
        tools: list[BetaToolUnionParam] = []
        extra_headers: dict[str, str] = {}
        for tool in model_request_parameters.builtin_tools:
            if isinstance(tool, WebSearchTool):
                user_location = UserLocation(type='approximate', **tool.user_location) if tool.user_location else None
                tools.append(
                    BetaWebSearchTool20250305Param(
                        name='web_search',
                        type='web_search_20250305',
                        max_uses=tool.max_uses,
                        allowed_domains=tool.allowed_domains,
                        blocked_domains=tool.blocked_domains,
                        user_location=user_location,
                    )
                )
            elif isinstance(tool, CodeExecutionTool):  # pragma: no branch
                extra_headers['anthropic-beta'] = 'code-execution-2025-05-22'
                tools.append(BetaCodeExecutionTool20250522Param(name='code_execution', type='code_execution_20250522'))
            else:  # pragma: no cover
                raise UserError(
                    f'`{tool.__class__.__name__}` is not supported by `AnthropicModel`. If it should be, please file an issue.'
                )
        return tools, extra_headers

    async def _map_message(self, messages: list[ModelMessage]) -> tuple[str, list[BetaMessageParam]]:  # noqa: C901
        """Just maps a `upsonic.Message` to a `anthropic.types.MessageParam`."""
        system_prompt_parts: list[str] = []
        anthropic_messages: list[BetaMessageParam] = []
        for m in messages:
            if isinstance(m, ModelRequest):
                user_content_params: list[BetaContentBlockParam] = []
                for request_part in m.parts:
                    if isinstance(request_part, SystemPromptPart):
                        system_prompt_parts.append(request_part.content)
                    elif isinstance(request_part, UserPromptPart):
                        async for content in self._map_user_prompt(request_part):
                            user_content_params.append(content)
                    elif isinstance(request_part, ToolReturnPart):
                        tool_result_block_param = BetaToolResultBlockParam(
                            tool_use_id=_guard_tool_call_id(t=request_part),
                            type='tool_result',
                            content=request_part.model_response_str(),
                            is_error=False,
                        )
                        user_content_params.append(tool_result_block_param)
                    elif isinstance(request_part, RetryPromptPart):  # pragma: no branch
                        if request_part.tool_name is None:
                            text = request_part.model_response()  # pragma: no cover
                            retry_param = BetaTextBlockParam(type='text', text=text)  # pragma: no cover
                        else:
                            retry_param = BetaToolResultBlockParam(
                                tool_use_id=_guard_tool_call_id(t=request_part),
                                type='tool_result',
                                content=request_part.model_response(),
                                is_error=True,
                            )
                        user_content_params.append(retry_param)
                if len(user_content_params) > 0:
                    anthropic_messages.append(BetaMessageParam(role='user', content=user_content_params))
            elif isinstance(m, ModelResponse):
                assistant_content_params: list[
                    BetaTextBlockParam
                    | BetaToolUseBlockParam
                    | BetaServerToolUseBlockParam
                    | BetaWebSearchToolResultBlockParam
                    | BetaCodeExecutionToolResultBlockParam
                    | BetaThinkingBlockParam
                    | BetaRedactedThinkingBlockParam
                ] = []
                for response_part in m.parts:
                    if isinstance(response_part, TextPart):
                        if response_part.content:
                            assistant_content_params.append(BetaTextBlockParam(text=response_part.content, type='text'))
                    elif isinstance(response_part, ToolCallPart):
                        tool_use_block_param = BetaToolUseBlockParam(
                            id=_guard_tool_call_id(t=response_part),
                            type='tool_use',
                            name=response_part.tool_name,
                            input=response_part.args_as_dict(),
                        )
                        assistant_content_params.append(tool_use_block_param)
                    elif isinstance(response_part, ThinkingPart):
                        if (
                            response_part.provider_name == self.system and response_part.signature is not None
                        ):  # pragma: no branch
                            if response_part.id == 'redacted_thinking':
                                assistant_content_params.append(
                                    BetaRedactedThinkingBlockParam(
                                        data=response_part.signature,
                                        type='redacted_thinking',
                                    )
                                )
                            else:
                                assistant_content_params.append(
                                    BetaThinkingBlockParam(
                                        thinking=response_part.content,
                                        signature=response_part.signature,
                                        type='thinking',
                                    )
                                )
                        elif response_part.content:  # pragma: no branch
                            start_tag, end_tag = self.profile.thinking_tags
                            assistant_content_params.append(
                                BetaTextBlockParam(
                                    text='\n'.join([start_tag, response_part.content, end_tag]), type='text'
                                )
                            )
                    elif isinstance(response_part, BuiltinToolCallPart):
                        if response_part.provider_name == self.system:
                            tool_use_id = _guard_tool_call_id(t=response_part)
                            if response_part.tool_name == WebSearchTool.kind:
                                server_tool_use_block_param = BetaServerToolUseBlockParam(
                                    id=tool_use_id,
                                    type='server_tool_use',
                                    name='web_search',
                                    input=response_part.args_as_dict(),
                                )
                                assistant_content_params.append(server_tool_use_block_param)
                            elif response_part.tool_name == CodeExecutionTool.kind:  # pragma: no branch
                                server_tool_use_block_param = BetaServerToolUseBlockParam(
                                    id=tool_use_id,
                                    type='server_tool_use',
                                    name='code_execution',
                                    input=response_part.args_as_dict(),
                                )
                                assistant_content_params.append(server_tool_use_block_param)
                    elif isinstance(response_part, BuiltinToolReturnPart):
                        if response_part.provider_name == self.system:
                            tool_use_id = _guard_tool_call_id(t=response_part)
                            if response_part.tool_name in (
                                WebSearchTool.kind,
                                'web_search_tool_result',  # Backward compatibility
                            ) and isinstance(response_part.content, dict | list):
                                assistant_content_params.append(
                                    BetaWebSearchToolResultBlockParam(
                                        tool_use_id=tool_use_id,
                                        type='web_search_tool_result',
                                        content=cast(
                                            BetaWebSearchToolResultBlockParamContentParam,
                                            response_part.content,  # pyright: ignore[reportUnknownMemberType]
                                        ),
                                    )
                                )
                            elif response_part.tool_name in (  # pragma: no branch
                                CodeExecutionTool.kind,
                                'code_execution_tool_result',  # Backward compatibility
                            ) and isinstance(response_part.content, dict):
                                assistant_content_params.append(
                                    BetaCodeExecutionToolResultBlockParam(
                                        tool_use_id=tool_use_id,
                                        type='code_execution_tool_result',
                                        content=cast(
                                            BetaCodeExecutionToolResultBlockParamContentParam,
                                            response_part.content,  # pyright: ignore[reportUnknownMemberType]
                                        ),
                                    )
                                )
                    else:
                        assert_never(response_part)
                if len(assistant_content_params) > 0:
                    anthropic_messages.append(BetaMessageParam(role='assistant', content=assistant_content_params))
            else:
                assert_never(m)
        if instructions := self._get_instructions(messages):
            system_prompt_parts.insert(0, instructions)
        system_prompt = '\n\n'.join(system_prompt_parts)
        return system_prompt, anthropic_messages

    @staticmethod
    async def _map_user_prompt(
        part: UserPromptPart,
    ) -> AsyncGenerator[BetaContentBlockParam]:
        if isinstance(part.content, str):
            if part.content:  # Only yield non-empty text
                yield BetaTextBlockParam(text=part.content, type='text')
        else:
            for item in part.content:
                if isinstance(item, str):
                    if item:  # Only yield non-empty text
                        yield BetaTextBlockParam(text=item, type='text')
                elif isinstance(item, BinaryContent):
                    if item.is_image:
                        yield BetaImageBlockParam(
                            source={'data': io.BytesIO(item.data), 'media_type': item.media_type, 'type': 'base64'},  # type: ignore
                            type='image',
                        )
                    elif item.media_type == 'application/pdf':
                        yield BetaBase64PDFBlockParam(
                            source=BetaBase64PDFSourceParam(
                                data=io.BytesIO(item.data),
                                media_type='application/pdf',
                                type='base64',
                            ),
                            type='document',
                        )
                    else:
                        raise RuntimeError('Only images and PDFs are supported for binary content')
                elif isinstance(item, ImageUrl):
                    yield BetaImageBlockParam(source={'type': 'url', 'url': item.url}, type='image')
                elif isinstance(item, DocumentUrl):
                    if item.media_type == 'application/pdf':
                        yield BetaBase64PDFBlockParam(source={'url': item.url, 'type': 'url'}, type='document')
                    elif item.media_type == 'text/plain':
                        downloaded_item = await download_item(item, data_format='text')
                        yield BetaBase64PDFBlockParam(
                            source=BetaPlainTextSourceParam(
                                data=downloaded_item['data'], media_type=item.media_type, type='text'
                            ),
                            type='document',
                        )
                    else:  # pragma: no cover
                        raise RuntimeError(f'Unsupported media type: {item.media_type}')
                else:
                    raise RuntimeError(f'Unsupported content type: {type(item)}')  # pragma: no cover

    @staticmethod
    def _map_tool_definition(f: ToolDefinition) -> BetaToolParam:
        return {
            'name': f.name,
            'description': f.description or '',
            'input_schema': f.parameters_json_schema,
        }


def _map_usage(message: BetaMessage | BetaRawMessageStartEvent | BetaRawMessageDeltaEvent) -> usage.RequestUsage:
    if isinstance(message, BetaMessage):
        response_usage = message.usage
    elif isinstance(message, BetaRawMessageStartEvent):
        response_usage = message.message.usage
    elif isinstance(message, BetaRawMessageDeltaEvent):
        response_usage = message.usage
    else:
        assert_never(message)

    # Store all integer-typed usage values in the details, except 'output_tokens' which is represented exactly by
    # `response_tokens`
    details: dict[str, int] = {
        key: value for key, value in response_usage.model_dump().items() if isinstance(value, int)
    }

    # Usage coming from the RawMessageDeltaEvent doesn't have input token data, hence using `get`
    # Tokens are only counted once between input_tokens, cache_creation_input_tokens, and cache_read_input_tokens
    # This approach maintains request_tokens as the count of all input tokens, with cached counts as details
    cache_write_tokens = details.get('cache_creation_input_tokens', 0)
    cache_read_tokens = details.get('cache_read_input_tokens', 0)
    request_tokens = details.get('input_tokens', 0) + cache_write_tokens + cache_read_tokens

    return usage.RequestUsage(
        input_tokens=request_tokens,
        cache_read_tokens=cache_read_tokens,
        cache_write_tokens=cache_write_tokens,
        output_tokens=response_usage.output_tokens,
        details=details,
    )


@dataclass
class AnthropicStreamedResponse(StreamedResponse):
    """Implementation of `StreamedResponse` for Anthropic models."""

    _model_name: AnthropicModelName
    _response: AsyncIterable[BetaRawMessageStreamEvent]
    _timestamp: datetime
    _provider_name: str

    async def _get_event_iterator(self) -> AsyncIterator[ModelResponseStreamEvent]:  # noqa: C901
        current_block: BetaContentBlock | None = None

        async for event in self._response:
            if isinstance(event, BetaRawMessageStartEvent):
                self._usage = _map_usage(event)
                self.provider_response_id = event.message.id

            elif isinstance(event, BetaRawContentBlockStartEvent):
                current_block = event.content_block
                if isinstance(current_block, BetaTextBlock) and current_block.text:
                    maybe_event = self._parts_manager.handle_text_delta(
                        vendor_part_id=event.index, content=current_block.text
                    )
                    if maybe_event is not None:  # pragma: no branch
                        yield maybe_event
                elif isinstance(current_block, BetaThinkingBlock):
                    yield self._parts_manager.handle_thinking_delta(
                        vendor_part_id=event.index,
                        content=current_block.thinking,
                        signature=current_block.signature,
                        provider_name=self.provider_name,
                    )
                elif isinstance(current_block, BetaRedactedThinkingBlock):
                    yield self._parts_manager.handle_thinking_delta(
                        vendor_part_id=event.index,
                        id='redacted_thinking',
                        signature=current_block.data,
                        provider_name=self.provider_name,
                    )
                elif isinstance(current_block, BetaToolUseBlock):
                    maybe_event = self._parts_manager.handle_tool_call_delta(
                        vendor_part_id=event.index,
                        tool_name=current_block.name,
                        args=cast(dict[str, Any], current_block.input) or None,
                        tool_call_id=current_block.id,
                    )
                    if maybe_event is not None:  # pragma: no branch
                        yield maybe_event
                elif isinstance(current_block, BetaServerToolUseBlock):
                    yield self._parts_manager.handle_builtin_tool_call_part(
                        vendor_part_id=event.index,
                        part=_map_server_tool_use_block(current_block, self.provider_name),
                    )
                elif isinstance(current_block, BetaWebSearchToolResultBlock):
                    yield self._parts_manager.handle_builtin_tool_return_part(
                        vendor_part_id=event.index,
                        part=_map_web_search_tool_result_block(current_block, self.provider_name),
                    )
                elif isinstance(current_block, BetaCodeExecutionToolResultBlock):
                    yield self._parts_manager.handle_builtin_tool_return_part(
                        vendor_part_id=event.index,
                        part=_map_code_execution_tool_result_block(current_block, self.provider_name),
                    )

            elif isinstance(event, BetaRawContentBlockDeltaEvent):
                if isinstance(event.delta, BetaTextDelta):
                    maybe_event = self._parts_manager.handle_text_delta(
                        vendor_part_id=event.index, content=event.delta.text
                    )
                    if maybe_event is not None:  # pragma: no branch
                        yield maybe_event
                elif isinstance(event.delta, BetaThinkingDelta):
                    yield self._parts_manager.handle_thinking_delta(
                        vendor_part_id=event.index,
                        content=event.delta.thinking,
                        provider_name=self.provider_name,
                    )
                elif isinstance(event.delta, BetaSignatureDelta):
                    yield self._parts_manager.handle_thinking_delta(
                        vendor_part_id=event.index,
                        signature=event.delta.signature,
                        provider_name=self.provider_name,
                    )
                elif isinstance(event.delta, BetaInputJSONDelta):
                    maybe_event = self._parts_manager.handle_tool_call_delta(
                        vendor_part_id=event.index,
                        args=event.delta.partial_json,
                    )
                    if maybe_event is not None:  # pragma: no branch
                        yield maybe_event
                # TODO: We need to handle citations.
                elif isinstance(event.delta, BetaCitationsDelta):
                    pass

            elif isinstance(event, BetaRawMessageDeltaEvent):
                self._usage = _map_usage(event)
                if raw_finish_reason := event.delta.stop_reason:  # pragma: no branch
                    self.provider_details = {'finish_reason': raw_finish_reason}
                    self.finish_reason = _FINISH_REASON_MAP.get(raw_finish_reason)

            elif isinstance(event, BetaRawContentBlockStopEvent | BetaRawMessageStopEvent):  # pragma: no branch
                current_block = None

    @property
    def model_name(self) -> AnthropicModelName:
        """Get the model name of the response."""
        return self._model_name

    @property
    def provider_name(self) -> str:
        """Get the provider name."""
        return self._provider_name

    @property
    def timestamp(self) -> datetime:
        """Get the timestamp of the response."""
        return self._timestamp


def _map_server_tool_use_block(item: BetaServerToolUseBlock, provider_name: str) -> BuiltinToolCallPart:
    if item.name == 'web_search':
        return BuiltinToolCallPart(
            provider_name=provider_name,
            tool_name=WebSearchTool.kind,
            args=cast(dict[str, Any], item.input) or None,
            tool_call_id=item.id,
        )
    elif item.name == 'code_execution':
        return BuiltinToolCallPart(
            provider_name=provider_name,
            tool_name=CodeExecutionTool.kind,
            args=cast(dict[str, Any], item.input) or None,
            tool_call_id=item.id,
        )
    else:
        assert_never(item.name)


web_search_tool_result_content_ta: TypeAdapter[BetaWebSearchToolResultBlockContent] = TypeAdapter(
    BetaWebSearchToolResultBlockContent
)


def _map_web_search_tool_result_block(item: BetaWebSearchToolResultBlock, provider_name: str) -> BuiltinToolReturnPart:
    return BuiltinToolReturnPart(
        provider_name=provider_name,
        tool_name=WebSearchTool.kind,
        content=web_search_tool_result_content_ta.dump_python(item.content, mode='json'),
        tool_call_id=item.tool_use_id,
    )


code_execution_tool_result_content_ta: TypeAdapter[BetaCodeExecutionToolResultBlockContent] = TypeAdapter(
    BetaCodeExecutionToolResultBlockContent
)


def _map_code_execution_tool_result_block(
    item: BetaCodeExecutionToolResultBlock, provider_name: str
) -> BuiltinToolReturnPart:
    return BuiltinToolReturnPart(
        provider_name=provider_name,
        tool_name=CodeExecutionTool.kind,
        content=code_execution_tool_result_content_ta.dump_python(item.content, mode='json'),
        tool_call_id=item.tool_use_id,
    )


from .async_utils import AsyncExecutionMixin
from .printing import (
    print_price_id_summary, 
    call_end,
    get_estimated_cost,
    get_estimated_cost_from_usage,
    get_estimated_cost_from_run_result,
    get_estimated_cost_from_stream_result,
    get_estimated_cost_from_agent
)

__all__ = [
    "AsyncExecutionMixin",
    "print_price_id_summary",
    "call_end",
    "get_estimated_cost",
    "get_estimated_cost_from_usage",
    "get_estimated_cost_from_run_result",
    "get_estimated_cost_from_stream_result",
    "get_estimated_cost_from_agent",
]


from __future__ import annotations
from typing import TYPE_CHECKING, Any

from .base import (
    BaseVectorDBProvider,
)

from .config import (
    DistanceMetric,
    IndexType,
    QuantizationConfig,
)

if TYPE_CHECKING:
    from .providers.chroma import ChromaProvider
    from .providers.faiss import FaissProvider
    from .providers.pinecone import PineconeProvider
    from .providers.qdrant import QdrantProvider
    from .providers.milvus import MilvusProvider
    from .providers.weaviate import WeaviateProvider
    from .providers.pgvector import PgvectorProvider

def _get_provider_classes():
    """Get provider classes with lazy importing."""
    providers = {}
    
    try:
        from .providers.chroma import ChromaProvider
        providers['ChromaProvider'] = ChromaProvider
    except ImportError:
        pass
    
    try:
        from .providers.faiss import FaissProvider
        providers['FaissProvider'] = FaissProvider
    except ImportError:
        pass
    
    try:
        from .providers.pinecone import PineconeProvider
        providers['PineconeProvider'] = PineconeProvider
    except ImportError:
        pass
    
    try:
        from .providers.qdrant import QdrantProvider
        providers['QdrantProvider'] = QdrantProvider
    except ImportError:
        pass
    
    try:
        from .providers.milvus import MilvusProvider
        providers['MilvusProvider'] = MilvusProvider
    except ImportError:
        pass
    
    try:
        from .providers.weaviate import WeaviateProvider
        providers['WeaviateProvider'] = WeaviateProvider
    except ImportError:
        pass
    
    try:
        from .providers.pgvector import PgvectorProvider
        providers['PgvectorProvider'] = PgvectorProvider
    except ImportError:
        pass
    
    return providers


def __getattr__(name: str) -> Any:
    """Lazy import of provider classes."""
    providers = _get_provider_classes()
    if name in providers:
        return providers[name]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


__all__ = [
    'ChromaProvider',
    'FaissProvider',
    'PineconeProvider',
    'QdrantProvider',
    'MilvusProvider',
    'WeaviateProvider',
    'PgvectorProvider',
]





import unittest
import uuid
from upsonic.text_splitter.character import CharacterChunker, CharacterChunkingConfig
from upsonic.schemas.data_models import Document, Chunk


class TestCharacterChunkingSimple(unittest.TestCase):
    """Simplified tests for CharacterChunker that match the actual implementation."""

    def setUp(self):
        """Set up test documents."""
        self.basic_doc = Document(
            content="This is the first paragraph.\n\nThis is the second paragraph.\n\nThis is the third paragraph.",
            metadata={'source': 'test', 'type': 'basic'},
            document_id=str(uuid.uuid4())
        )
        
        self.long_doc = Document(
            content="This is paragraph 1.\n\nThis is paragraph 2.\n\nThis is paragraph 3.\n\nThis is paragraph 4.\n\nThis is paragraph 5.",
            metadata={'source': 'test', 'type': 'long'},
            document_id=str(uuid.uuid4())
        )

    def test_basic_chunking(self):
        """Test basic chunking functionality."""
        config = CharacterChunkingConfig(chunk_size=100, chunk_overlap=20)
        chunker = CharacterChunker(config)
        
        chunks = chunker.chunk([self.basic_doc])
        
        self.assertGreater(len(chunks), 0)
        self.assertTrue(all(isinstance(chunk, Chunk) for chunk in chunks))
        self.assertTrue(all(chunk.text_content.strip() for chunk in chunks))

    def test_different_separators(self):
        """Test different separator configurations."""
        # Test with paragraph separator
        config = CharacterChunkingConfig(separator="\n\n", chunk_size=50)
        chunker = CharacterChunker(config)
        chunks = chunker.chunk([self.basic_doc])
        
        self.assertGreater(len(chunks), 0)
        
        # Test with regex separator
        config_regex = CharacterChunkingConfig(
            separator=r"\n+", 
            is_separator_regex=True,
            chunk_size=50
        )
        chunker_regex = CharacterChunker(config_regex)
        chunks_regex = chunker_regex.chunk([self.basic_doc])
        
        self.assertGreater(len(chunks_regex), 0)

    def test_empty_document(self):
        """Test handling of empty documents."""
        empty_doc = Document(
            content="",
            metadata={'source': 'empty'},
            document_id=str(uuid.uuid4())
        )
        
        config = CharacterChunkingConfig()
        chunker = CharacterChunker(config)
        chunks = chunker.chunk([empty_doc])
        
        self.assertEqual(len(chunks), 0)

    def test_chunk_overlap(self):
        """Test chunk overlap functionality."""
        config = CharacterChunkingConfig(chunk_size=50, chunk_overlap=10)
        chunker = CharacterChunker(config)
        
        chunks = chunker.chunk([self.long_doc])
        
        self.assertGreater(len(chunks), 1)
        # Check that chunks have proper metadata
        for chunk in chunks:
            self.assertEqual(chunk.document_id, self.long_doc.document_id)
            self.assertIn('source', chunk.metadata)

    def test_keep_separator(self):
        """Test separator retention functionality."""
        # Test with keep_separator=True
        config_keep = CharacterChunkingConfig(
            separator="\n\n",
            keep_separator=True,
            chunk_size=50
        )
        chunker_keep = CharacterChunker(config_keep)
        chunks_keep = chunker_keep.chunk([self.basic_doc])
        
        # Test with keep_separator=False
        config_no_keep = CharacterChunkingConfig(
            separator="\n\n",
            keep_separator=False,
            chunk_size=50
        )
        chunker_no_keep = CharacterChunker(config_no_keep)
        chunks_no_keep = chunker_no_keep.chunk([self.basic_doc])
        
        self.assertGreater(len(chunks_keep), 0)
        self.assertGreater(len(chunks_no_keep), 0)


if __name__ == "__main__":
    unittest.main()


import asyncio
import re
import csv
import io
from pathlib import Path
from typing import List, Optional, Tuple, Union, Any
import concurrent.futures

from upsonic.schemas.data_models import Document
from upsonic.loaders.base import BaseLoader
from upsonic.loaders.config import PdfPlumberLoaderConfig

try:
    import pdfplumber
    _PDFPLUMBER_AVAILABLE = True
except ImportError:
    pdfplumber = None
    _PDFPLUMBER_AVAILABLE = False


try:
    from rapidocr_onnxruntime import RapidOCR
    OCR_ENGINE = RapidOCR()
    _RAPIDOCR_AVAILABLE = True
except ImportError:
    RapidOCR = None
    OCR_ENGINE = None
    _RAPIDOCR_AVAILABLE = False




class PdfPlumberLoader(BaseLoader):
    """
    An advanced, high-performance loader for PDF documents using pdfplumber.

    This loader is specifically designed for AI agent frameworks, providing:
    - Superior table extraction and preservation
    - Advanced layout analysis and preservation
    - Character-level text extraction precision
    - OCR support for scanned documents (hybrid mode)
    - Complex document structure handling
    - High-quality extraction for RAG applications

    pdfplumber excels at structured documents with tables, forms, and complex
    layouts, making it ideal for enterprise document processing in AI systems.
    """

    def __init__(self, config: PdfPlumberLoaderConfig):
        """
        Initializes the PdfPlumberLoader with its specific configuration.

        Args:
            config: A PdfPlumberLoaderConfig object with settings for PDF processing.
        """
        if not _PDFPLUMBER_AVAILABLE:
            from upsonic.utils.printing import import_error
            import_error(
                package_name="pdfplumber",
                install_command='pip install "upsonic[loaders]"',
                feature_name="pdfplumber PDF loader"
            )
        if config.extraction_mode in ("ocr_only", "hybrid") and not _RAPIDOCR_AVAILABLE:
            from upsonic.utils.printing import import_error
            import_error(
                package_name="rapidocr-onnxruntime",
                install_command='pip install "upsonic[loaders]"',
                feature_name="PDF OCR functionality"
            )
        super().__init__(config)
        self.config: PdfPlumberLoaderConfig = config


    @classmethod
    def get_supported_extensions(cls) -> List[str]:
        """Gets a list of file extensions supported by this loader."""
        return [".pdf"]

    def load(self, source: Union[str, Path, List[Union[str, Path]]]) -> List[Document]:
        """
        Loads all PDF documents from the given source synchronously.

        This is a convenience wrapper around the async `aload` method.
        """
        try:
            asyncio.get_running_loop()
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(asyncio.run, self.aload(source))
                return future.result()
        except RuntimeError:
            return asyncio.run(self.aload(source))

    async def aload(self, source: Union[str, Path, List[Union[str, Path]]]) -> List[Document]:
        """
        Loads all PDF documents from the given source asynchronously and concurrently.
        """
        pdf_paths = self._resolve_sources(source)
        if not pdf_paths:
            return []

        tasks = [self._process_single_pdf(path) for path in pdf_paths]
        results = await asyncio.gather(*tasks, return_exceptions=True)

        documents = []
        for result in results:
            if isinstance(result, Exception):
                if self.config.error_handling == "raise":
                    raise result
                elif self.config.error_handling == "warn":
                    self._logger.warning("Failed to process PDF: %s", result)
            else:
                documents.extend(result)

        return documents
    
    def batch(self, sources: List[Union[str, Path]]) -> List[Document]:
        """Loads documents from a list of sources, leveraging the core `load` method."""
        return self.load(sources)

    async def abatch(self, sources: List[Union[str, Path]]) -> List[Document]:
        """Loads documents from a list of sources asynchronously, leveraging the core `aload` method."""
        return await self.aload(sources)



    async def _process_single_pdf(self, path: Path) -> List[Document]:
        """
        Processes a single PDF file, consolidating all page content into a single Document.
        
        This method orchestrates the entire processing pipeline:
        1. Document ID generation and duplicate checking
        2. File size validation
        3. PDF opening and password handling
        4. Page-by-page content extraction
        5. Table extraction and formatting
        6. OCR for scanned content (if configured)
        7. Content cleaning and normalization
        8. Metadata creation
        """
        try:
            document_id = self._generate_document_id(path)
            if document_id in self._processed_document_ids:
                raise FileExistsError(
                    f"Source file '{path.resolve()}' has already been processed by this loader instance."
                )
            self._processed_document_ids.add(document_id)
            
            if not self._check_file_size(path):
                return []
            
            # Open PDF with pdfplumber
            pdf_handle = await asyncio.to_thread(
                pdfplumber.open, 
                str(path),
                password=self.config.pdf_password
            )

            try:
                pages = pdf_handle.pages
                
                # Determine page range
                start_idx = (self.config.start_page - 1) if self.config.start_page else 0
                end_idx = self.config.end_page if self.config.end_page else len(pages)
                pages_to_process = pages[start_idx:end_idx]

                # Process pages concurrently
                page_tasks = [
                    self._extract_page_content(page, page_num=start_idx + i + 1) 
                    for i, page in enumerate(pages_to_process)
                ]
                page_contents_with_nums = await asyncio.gather(*page_tasks)
                
                page_contents = [content for content, _num in page_contents_with_nums]

                # Apply content cleaning
                if self.config.extra_whitespace_removal:
                    page_contents = [self._normalize_whitespace(content) for content in page_contents]

                if self.config.clean_page_numbers:
                    page_contents, _ = self._clean_page_numbers(page_contents)
                
                # Consolidate into single document
                full_content = "\n\n".join(page_contents).strip()

                if self.config.skip_empty_content and not full_content:
                    return []
                
                # Create metadata
                metadata = self._create_metadata(path)
                metadata["page_count"] = len(page_contents)
                
                # Add pdfplumber-specific metadata
                if self.config.include_metadata and pages_to_process:
                    first_page = pages_to_process[0]
                    if self.config.extract_page_dimensions:
                        metadata["page_width"] = first_page.width
                        metadata["page_height"] = first_page.height
                    
                    # Add PDF metadata if available
                    if hasattr(pdf_handle, 'metadata') and pdf_handle.metadata:
                        pdf_metadata = pdf_handle.metadata
                        if pdf_metadata.get('Title'):
                            metadata["pdf_title"] = pdf_metadata['Title']
                        if pdf_metadata.get('Author'):
                            metadata["pdf_author"] = pdf_metadata['Author']
                        if pdf_metadata.get('Subject'):
                            metadata["pdf_subject"] = pdf_metadata['Subject']
                        if pdf_metadata.get('Creator'):
                            metadata["pdf_creator"] = pdf_metadata['Creator']
                
                if self.config.include_metadata:
                    return [Document(document_id=document_id, content=full_content, metadata=metadata)]
                else:
                    return [Document(document_id=document_id, content=full_content)]

            finally:
                # Always close the PDF handle
                await asyncio.to_thread(pdf_handle.close)

        except (PermissionError, FileExistsError) as e:
            raise e
        except Exception as e:
            return self._handle_loading_error(str(path), e)


    async def _extract_page_content(self, page, page_num: int) -> Tuple[str, int]:
        """
        Extracts content from a single page based on the `extraction_mode`.
        
        This method leverages pdfplumber's advanced features:
        - Text extraction with layout preservation
        - Table detection and extraction
        - Image information extraction
        - Character-level precision
        - OCR fallback for scanned content
        
        Args:
            page: A pdfplumber page object
            page_num: The page number (1-indexed)
            
        Returns:
            Tuple of (extracted_content, page_num)
        """
        content_parts = []
        
        # Apply crop box if specified
        working_page = page
        if self.config.crop_box:
            working_page = await asyncio.to_thread(
                page.crop, 
                self.config.crop_box
            )

        # Extract text based on mode
        if self.config.extraction_mode in ("text_only", "hybrid"):
            text_content = await self._extract_text_from_page(working_page)
            if text_content:
                content_parts.append(text_content)
        
        # Extract tables if configured
        if self.config.extract_tables and self.config.extraction_mode in ("text_only", "hybrid"):
            tables_content = await self._extract_tables_from_page(working_page)
            if tables_content:
                content_parts.append(tables_content)
        
        # Extract image information if configured
        if self.config.extract_images:
            images_info = await self._extract_images_info(working_page)
            if images_info:
                content_parts.append(images_info)
        
        # Extract annotations if configured
        if self.config.extract_annotations:
            annotations = await self._extract_annotations(working_page)
            if annotations:
                content_parts.append(annotations)
        
        # Perform OCR if needed
        if self.config.extraction_mode in ("ocr_only", "hybrid"):
            ocr_text = await self._perform_ocr(working_page)
            if ocr_text:
                content_parts.append(ocr_text)
        
        full_content = "\n\n".join(content_parts).strip()
        return full_content, page_num


    async def _extract_text_from_page(self, page) -> str:
        """
        Extracts text from a page using pdfplumber's advanced text extraction.
        
        Uses configuration settings for:
        - Layout preservation
        - Character/word/line margins
        - Text flow analysis
        """
        extract_text_kwargs = {
            "x_tolerance": self.config.char_margin,
            "y_tolerance": self.config.line_margin,
            "layout": self.config.layout_mode == "layout",
            "keep_blank_chars": self.config.keep_blank_chars,
        }
        
        if self.config.use_text_flow:
            extract_text_kwargs["use_text_flow"] = True
        
        text = await asyncio.to_thread(
            page.extract_text,
            **extract_text_kwargs
        )
        
        return text or ""


    async def _extract_tables_from_page(self, page) -> str:
        """
        Extracts and formats tables from a page using pdfplumber's superior table detection.
        
        pdfplumber excels at table extraction, making it ideal for structured documents.
        Tables are formatted according to the configured table_format.
        """
        tables = await asyncio.to_thread(
            page.extract_tables,
            table_settings=self.config.table_settings
        )
        
        if not tables:
            return ""
        
        formatted_tables = []
        for i, table in enumerate(tables):
            if not table or not any(table):
                continue
            
            table_title = f"\n[Table {i+1}]\n"
            formatted_table = self._format_table(table, self.config.table_format)
            formatted_tables.append(table_title + formatted_table)
        
        return "\n\n".join(formatted_tables)


    def _format_table(self, table: List[List[Any]], format_type: str) -> str:
        """
        Formats a table according to the specified format type.
        
        Args:
            table: 2D list representing the table
            format_type: One of "text", "markdown", "csv", "grid"
            
        Returns:
            Formatted table as a string
        """
        if not table:
            return ""
        
        # Clean None values and convert to strings
        cleaned_table = [
            [str(cell) if cell is not None else "" for cell in row]
            for row in table
        ]
        
        if format_type == "markdown":
            return self._format_table_markdown(cleaned_table)
        elif format_type == "csv":
            return self._format_table_csv(cleaned_table)
        elif format_type == "grid":
            return self._format_table_grid(cleaned_table)
        else:  # "text"
            return self._format_table_text(cleaned_table)


    def _format_table_markdown(self, table: List[List[str]]) -> str:
        """Formats a table as Markdown (ideal for AI agents and RAG)."""
        if not table or len(table) < 1:
            return ""
        
        lines = []
        
        # Header row
        if table:
            header = table[0]
            lines.append("| " + " | ".join(header) + " |")
            lines.append("|" + "|".join(["---" for _ in header]) + "|")
            
            # Data rows
            for row in table[1:]:
                # Pad row if it has fewer columns than header
                padded_row = row + [""] * (len(header) - len(row))
                lines.append("| " + " | ".join(padded_row[:len(header)]) + " |")
        
        return "\n".join(lines)


    def _format_table_csv(self, table: List[List[str]]) -> str:
        """Formats a table as CSV."""
        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerows(table)
        return output.getvalue().strip()


    def _format_table_grid(self, table: List[List[str]]) -> str:
        """Formats a table as ASCII grid."""
        if not table:
            return ""
        
        # Calculate column widths
        col_widths = []
        max_cols = max(len(row) for row in table)
        
        for col_idx in range(max_cols):
            max_width = max(
                len(row[col_idx]) if col_idx < len(row) else 0
                for row in table
            )
            col_widths.append(max(max_width, 3))
        
        # Build grid
        lines = []
        separator = "+" + "+".join(["-" * (w + 2) for w in col_widths]) + "+"
        
        for row in table:
            padded_row = row + [""] * (max_cols - len(row))
            line = "|" + "|".join([
                f" {cell:<{col_widths[i]}} "
                for i, cell in enumerate(padded_row)
            ]) + "|"
            lines.append(separator)
            lines.append(line)
        lines.append(separator)
        
        return "\n".join(lines)


    def _format_table_text(self, table: List[List[str]]) -> str:
        """Formats a table as plain text with spacing."""
        if not table:
            return ""
        
        # Calculate column widths
        max_cols = max(len(row) for row in table)
        col_widths = []
        
        for col_idx in range(max_cols):
            max_width = max(
                len(row[col_idx]) if col_idx < len(row) else 0
                for row in table
            )
            col_widths.append(max_width)
        
        # Format rows
        lines = []
        for row in table:
            padded_row = row + [""] * (max_cols - len(row))
            line = "  ".join([
                cell.ljust(col_widths[i])
                for i, cell in enumerate(padded_row)
            ])
            lines.append(line)
        
        return "\n".join(lines)


    async def _extract_images_info(self, page) -> str:
        """
        Extracts information about images on the page.
        
        Returns a formatted string with image metadata.
        """
        images = await asyncio.to_thread(lambda: page.images)
        
        if not images:
            return ""
        
        info_parts = [f"\n[Images: {len(images)} found on page]"]
        
        for i, img in enumerate(images):
            img_info = f"Image {i+1}: "
            if 'width' in img and 'height' in img:
                img_info += f"{img['width']}x{img['height']}px"
            if 'x0' in img and 'y0' in img:
                img_info += f" at ({img['x0']:.1f}, {img['y0']:.1f})"
            info_parts.append(img_info)
        
        return "\n".join(info_parts)


    async def _extract_annotations(self, page) -> str:
        """
        Extracts annotations and hyperlinks from the page.
        """
        annotations = await asyncio.to_thread(lambda: page.annots)
        hyperlinks = await asyncio.to_thread(lambda: page.hyperlinks)
        
        parts = []
        
        if annotations:
            parts.append(f"\n[Annotations: {len(annotations)}]")
            for i, annot in enumerate(annotations[:10]):  # Limit to 10
                if 'contents' in annot and annot['contents']:
                    parts.append(f"Note {i+1}: {annot['contents']}")
        
        if hyperlinks:
            parts.append(f"\n[Hyperlinks: {len(hyperlinks)}]")
            for i, link in enumerate(hyperlinks[:10]):  # Limit to 10
                if 'uri' in link:
                    parts.append(f"Link {i+1}: {link['uri']}")
        
        return "\n".join(parts) if parts else ""


    async def _perform_ocr(self, page) -> str:
        """
        Performs OCR on the page by converting it to an image.
        
        This is used when extraction_mode is "ocr_only" or "hybrid" (as fallback).
        For hybrid mode, OCR is performed on the entire page when text extraction yields little content.
        """
        try:
            # Convert page to image using pdfplumber's built-in functionality
            img = await asyncio.to_thread(
                page.to_image,
                resolution=150  # Good balance between quality and performance
            )
            
            # Get PIL image
            pil_image = await asyncio.to_thread(lambda: img.original)
            
            # Convert to bytes for OCR
            img_bytes_io = io.BytesIO()
            await asyncio.to_thread(pil_image.save, img_bytes_io, format='PNG')
            img_bytes = img_bytes_io.getvalue()
            
            # Run OCR
            loop = asyncio.get_running_loop()
            with concurrent.futures.ThreadPoolExecutor() as pool:
                ocr_result = await loop.run_in_executor(
                    pool, 
                    self._run_single_ocr, 
                    img_bytes
                )
            
            return ocr_result
            
        except (ImportError, ValueError, OSError) as e:
            self._logger.warning("OCR failed for page: %s", e)
            return ""


    @staticmethod
    def _run_single_ocr(image_data: bytes) -> str:
        """Helper function that runs the synchronous OCR engine."""
        if not OCR_ENGINE:
            return ""
        result, _ = OCR_ENGINE(image_data)
        if result:
            return "\n".join([item[1] for item in result])
        return ""


    @staticmethod
    def _normalize_whitespace(text: str) -> str:
        """Collapses multiple spaces/newlines and trims."""
        text = re.sub(r"\s*\n\s*", "\n", text)
        text = re.sub(r"[ \t]+", " ", text)
        return text.strip()
    

    def _clean_page_numbers(self, page_content_list: List[str]) -> Tuple[List[str], Optional[int]]:
        """
        Identifies and removes sequential page numbers from the top or bottom of pages.
        
        This method uses intelligent pattern matching to detect page numbers and removes them
        while optionally adding custom markers based on configuration.
        
        Args:
            page_content_list: List of page contents as strings
            
        Returns:
            Tuple of (cleaned_page_list, starting_page_number)
        """
        page_number_regex = re.compile(r"^\s*(\d+)\s*$")

        def find_page_number(content):
            lines = content.strip().split('\n')
            if not lines: 
                return None
            # Check first and last lines for standalone page numbers
            for line in [lines[0], lines[-1]]:
                match = page_number_regex.search(line)
                if match:
                    return int(match.group(1))
            return None

        page_numbers = [find_page_number(content) for content in page_content_list]
        
        # Find the best sequence match
        best_match, best_correct_count, best_shift = None, 0, None
        start_page = self.config.start_page or 1
        
        for shift in range(start_page, start_page + len(page_content_list) + 1):
            expected_numbers = list(range(shift, shift + len(page_numbers)))
            correct_count = sum(1 for actual, expected in zip(page_numbers, expected_numbers) if actual == expected)
            if correct_count > best_correct_count:
                best_correct_count, best_match, best_shift = correct_count, expected_numbers, shift
        
        # Apply cleaning if we found a good match (40% threshold)
        if best_match and best_correct_count / len(page_numbers) >= 0.4:
            cleaned_pages = []
            for i, content in enumerate(page_content_list):
                lines = content.strip().split('\n')
                expected_number = str(best_match[i])
                
                # Remove page number from top
                if lines and lines[0].strip() == expected_number:
                    lines.pop(0)
                # Remove page number from bottom
                if lines and lines[-1].strip() == expected_number:
                    lines.pop(-1)
                
                cleaned_content = '\n'.join(lines)

                # Add custom markers if configured
                start_marker = self.config.page_num_start_format.format(page_nr=expected_number) + "\n" if self.config.page_num_start_format else ""
                end_marker = "\n" + self.config.page_num_end_format.format(page_nr=expected_number) if self.config.page_num_end_format else ""

                cleaned_pages.append(f"{start_marker}{cleaned_content}{end_marker}")
            return cleaned_pages, best_shift
        
        return page_content_list, None




from __future__ import annotations
from typing import Optional, Type, Union, Dict, Any, List, Literal, Generic, TypeVar
from pydantic import BaseModel

from ..storage.base import Storage
from ..storage.memory.memory import Memory
from ..storage.providers import (
    InMemoryStorage,
    JSONStorage,
    Mem0Storage,
    PostgresStorage,
    RedisStorage,
    SqliteStorage,
    MongoStorage
)
from ..models import Model

# Generic type variable for storage providers
StorageType = TypeVar('StorageType', bound=Storage)


class DatabaseBase(Generic[StorageType]):
    """
    Base class for all database classes that combine storage providers with memory.
    
    This class provides a common interface and type safety for all database implementations.
    It uses generic types to ensure type safety while allowing different storage backends.
    """
    
    def __init__(
        self,
        storage: StorageType,
        memory: Memory
    ):
        """
        Initialize the database with storage and memory components.
        
        Args:
            storage: The storage provider instance
            memory: The memory instance
        """
        self.storage = storage
        self.memory = memory
    
    def __repr__(self) -> str:
        """String representation of the database instance."""
        return f"{self.__class__.__name__}(storage={type(self.storage).__name__}, memory={type(self.memory).__name__})"


class SqliteDatabase(DatabaseBase[SqliteStorage]):
    """
    Database class combining SqliteStorage and Memory attributes.
    """
    
    def __init__(
        self,
        # SqliteStorage attributes
        sessions_table_name: str,
        profiles_table_name: str,
        db_file: Optional[str] = None,
        # Memory attributes
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
        full_session_memory: bool = False,
        summary_memory: bool = False,
        user_analysis_memory: bool = False,
        user_profile_schema: Optional[Type[BaseModel]] = None,
        dynamic_user_profile: bool = False,
        num_last_messages: Optional[int] = None,
        model_provider: Optional[Union[Model, str]] = None,
        debug: bool = False,
        feed_tool_call_results: bool = False,
        user_memory_mode: Literal['update', 'replace'] = 'update'
    ):
        # Initialize storage
        storage = SqliteStorage(
            sessions_table_name=sessions_table_name,
            profiles_table_name=profiles_table_name,
            db_file=db_file
        )
        
        # Initialize memory
        memory = Memory(
            storage=storage,
            session_id=session_id,
            user_id=user_id,
            full_session_memory=full_session_memory,
            summary_memory=summary_memory,
            user_analysis_memory=user_analysis_memory,
            user_profile_schema=user_profile_schema,
            dynamic_user_profile=dynamic_user_profile,
            num_last_messages=num_last_messages,
            model_provider=model_provider,
            debug=debug,
            feed_tool_call_results=feed_tool_call_results,
            user_memory_mode=user_memory_mode
        )
        
        # Call parent constructor
        super().__init__(storage=storage, memory=memory)


class PostgresDatabase(DatabaseBase[PostgresStorage]):
    """
    Database class combining PostgresStorage and Memory attributes.
    """
    
    def __init__(
        self,
        # PostgresStorage attributes
        sessions_table_name: str,
        profiles_table_name: str,
        db_url: str,
        schema: str = "public",
        # Memory attributes
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
        full_session_memory: bool = False,
        summary_memory: bool = False,
        user_analysis_memory: bool = False,
        user_profile_schema: Optional[Type[BaseModel]] = None,
        dynamic_user_profile: bool = False,
        num_last_messages: Optional[int] = None,
        model_provider: Optional[Union[Model, str]] = None,
        debug: bool = False,
        feed_tool_call_results: bool = False,
        user_memory_mode: Literal['update', 'replace'] = 'update'
    ):
        # Initialize storage
        storage = PostgresStorage(
            sessions_table_name=sessions_table_name,
            profiles_table_name=profiles_table_name,
            db_url=db_url,
            schema=schema
        )
        
        # Initialize memory
        memory = Memory(
            storage=storage,
            session_id=session_id,
            user_id=user_id,
            full_session_memory=full_session_memory,
            summary_memory=summary_memory,
            user_analysis_memory=user_analysis_memory,
            user_profile_schema=user_profile_schema,
            dynamic_user_profile=dynamic_user_profile,
            num_last_messages=num_last_messages,
            model_provider=model_provider,
            debug=debug,
            feed_tool_call_results=feed_tool_call_results,
            user_memory_mode=user_memory_mode
        )
        
        # Call parent constructor
        super().__init__(storage=storage, memory=memory)


class MongoDatabase(DatabaseBase[MongoStorage]):
    """
    Database class combining MongoStorage and Memory attributes.
    """
    
    def __init__(
        self,
        # MongoStorage attributes
        db_url: str,
        database_name: str,
        sessions_collection_name: str = "interaction_sessions",
        profiles_collection_name: str = "user_profiles",
        # Memory attributes
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
        full_session_memory: bool = False,
        summary_memory: bool = False,
        user_analysis_memory: bool = False,
        user_profile_schema: Optional[Type[BaseModel]] = None,
        dynamic_user_profile: bool = False,
        num_last_messages: Optional[int] = None,
        model_provider: Optional[Union[Model, str]] = None,
        debug: bool = False,
        feed_tool_call_results: bool = False,
        user_memory_mode: Literal['update', 'replace'] = 'update'
    ):
        # Initialize storage
        storage = MongoStorage(
            db_url=db_url,
            database_name=database_name,
            sessions_collection_name=sessions_collection_name,
            profiles_collection_name=profiles_collection_name
        )
        
        # Initialize memory
        memory = Memory(
            storage=storage,
            session_id=session_id,
            user_id=user_id,
            full_session_memory=full_session_memory,
            summary_memory=summary_memory,
            user_analysis_memory=user_analysis_memory,
            user_profile_schema=user_profile_schema,
            dynamic_user_profile=dynamic_user_profile,
            num_last_messages=num_last_messages,
            model_provider=model_provider,
            debug=debug,
            feed_tool_call_results=feed_tool_call_results,
            user_memory_mode=user_memory_mode
        )
        
        # Call parent constructor
        super().__init__(storage=storage, memory=memory)


class RedisDatabase(DatabaseBase[RedisStorage]):
    """
    Database class combining RedisStorage and Memory attributes.
    """
    
    def __init__(
        self,
        # RedisStorage attributes
        prefix: str,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: Optional[str] = None,
        ssl: bool = False,
        expire: Optional[int] = None,
        # Memory attributes
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
        full_session_memory: bool = False,
        summary_memory: bool = False,
        user_analysis_memory: bool = False,
        user_profile_schema: Optional[Type[BaseModel]] = None,
        dynamic_user_profile: bool = False,
        num_last_messages: Optional[int] = None,
        model_provider: Optional[Union[Model, str]] = None,
        debug: bool = False,
        feed_tool_call_results: bool = False,
        user_memory_mode: Literal['update', 'replace'] = 'update'
    ):
        # Initialize storage
        storage = RedisStorage(
            prefix=prefix,
            host=host,
            port=port,
            db=db,
            password=password,
            ssl=ssl,
            expire=expire
        )
        
        # Initialize memory
        memory = Memory(
            storage=storage,
            session_id=session_id,
            user_id=user_id,
            full_session_memory=full_session_memory,
            summary_memory=summary_memory,
            user_analysis_memory=user_analysis_memory,
            user_profile_schema=user_profile_schema,
            dynamic_user_profile=dynamic_user_profile,
            num_last_messages=num_last_messages,
            model_provider=model_provider,
            debug=debug,
            feed_tool_call_results=feed_tool_call_results,
            user_memory_mode=user_memory_mode
        )
        
        # Call parent constructor
        super().__init__(storage=storage, memory=memory)


class InMemoryDatabase(DatabaseBase[InMemoryStorage]):
    """
    Database class combining InMemoryStorage and Memory attributes.
    """
    
    def __init__(
        self,
        # InMemoryStorage attributes
        max_sessions: Optional[int] = None,
        max_profiles: Optional[int] = None,
        # Memory attributes
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
        full_session_memory: bool = False,
        summary_memory: bool = False,
        user_analysis_memory: bool = False,
        user_profile_schema: Optional[Type[BaseModel]] = None,
        dynamic_user_profile: bool = False,
        num_last_messages: Optional[int] = None,
        model_provider: Optional[Union[Model, str]] = None,
        debug: bool = False,
        feed_tool_call_results: bool = False,
        user_memory_mode: Literal['update', 'replace'] = 'update'
    ):
        # Initialize storage
        storage = InMemoryStorage(
            max_sessions=max_sessions,
            max_profiles=max_profiles
        )
        
        # Initialize memory
        memory = Memory(
            storage=storage,
            session_id=session_id,
            user_id=user_id,
            full_session_memory=full_session_memory,
            summary_memory=summary_memory,
            user_analysis_memory=user_analysis_memory,
            user_profile_schema=user_profile_schema,
            dynamic_user_profile=dynamic_user_profile,
            num_last_messages=num_last_messages,
            model_provider=model_provider,
            debug=debug,
            feed_tool_call_results=feed_tool_call_results,
            user_memory_mode=user_memory_mode
        )
        
        # Call parent constructor
        super().__init__(storage=storage, memory=memory)


class JSONDatabase(DatabaseBase[JSONStorage]):
    """
    Database class combining JSONStorage and Memory attributes.
    """
    
    def __init__(
        self,
        # JSONStorage attributes
        directory_path: str,
        pretty_print: bool = True,
        # Memory attributes
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
        full_session_memory: bool = False,
        summary_memory: bool = False,
        user_analysis_memory: bool = False,
        user_profile_schema: Optional[Type[BaseModel]] = None,
        dynamic_user_profile: bool = False,
        num_last_messages: Optional[int] = None,
        model_provider: Optional[Union[Model, str]] = None,
        debug: bool = False,
        feed_tool_call_results: bool = False,
        user_memory_mode: Literal['update', 'replace'] = 'update'
    ):
        # Initialize storage
        storage = JSONStorage(
            directory_path=directory_path,
            pretty_print=pretty_print
        )
        
        # Initialize memory
        memory = Memory(
            storage=storage,
            session_id=session_id,
            user_id=user_id,
            full_session_memory=full_session_memory,
            summary_memory=summary_memory,
            user_analysis_memory=user_analysis_memory,
            user_profile_schema=user_profile_schema,
            dynamic_user_profile=dynamic_user_profile,
            num_last_messages=num_last_messages,
            model_provider=model_provider,
            debug=debug,
            feed_tool_call_results=feed_tool_call_results,
            user_memory_mode=user_memory_mode
        )
        
        # Call parent constructor
        super().__init__(storage=storage, memory=memory)


class Mem0Database(DatabaseBase[Mem0Storage]):
    """
    Database class combining Mem0Storage and Memory attributes.
    """
    
    def __init__(
        self,
        # Mem0Storage attributes
        api_key: Optional[str] = None,
        org_id: Optional[str] = None,
        project_id: Optional[str] = None,
        local_config: Optional[Dict[str, Any]] = None,
        namespace: str = "upsonic",
        infer: bool = False,
        custom_categories: Optional[List[str]] = None,
        enable_caching: bool = True,
        cache_ttl: int = 300,
        output_format: str = "v1.1",
        version: str = "v2",
        # Memory attributes
        session_id: Optional[str] = None,
        user_id: Optional[str] = None,
        full_session_memory: bool = False,
        summary_memory: bool = False,
        user_analysis_memory: bool = False,
        user_profile_schema: Optional[Type[BaseModel]] = None,
        dynamic_user_profile: bool = False,
        num_last_messages: Optional[int] = None,
        model_provider: Optional[Union[Model, str]] = None,
        debug: bool = False,
        feed_tool_call_results: bool = False,
        user_memory_mode: Literal['update', 'replace'] = 'update'
    ):
        # Initialize storage
        storage = Mem0Storage(
            api_key=api_key,
            org_id=org_id,
            project_id=project_id,
            local_config=local_config,
            namespace=namespace,
            infer=infer,
            custom_categories=custom_categories,
            enable_caching=enable_caching,
            cache_ttl=cache_ttl,
            output_format=output_format,
            version=version
        )
        
        # Initialize memory
        memory = Memory(
            storage=storage,
            session_id=session_id,
            user_id=user_id,
            full_session_memory=full_session_memory,
            summary_memory=summary_memory,
            user_analysis_memory=user_analysis_memory,
            user_profile_schema=user_profile_schema,
            dynamic_user_profile=dynamic_user_profile,
            num_last_messages=num_last_messages,
            model_provider=model_provider,
            debug=debug,
            feed_tool_call_results=feed_tool_call_results,
            user_memory_mode=user_memory_mode
        )
        
        # Call parent constructor
        super().__init__(storage=storage, memory=memory)



"""Built-in tools for AI models.

These tools are passed directly to the model provider's API and are not
executed by the Upsonic framework. They represent native capabilities
provided by the model providers themselves.
"""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Literal, Optional, TypedDict
try:
    import requests
    _REQUESTS_AVAILABLE = True
except ImportError:
    requests = None
    _REQUESTS_AVAILABLE = False


try:
    from bs4 import BeautifulSoup
    _BEAUTIFULSOUP_AVAILABLE = True
except ImportError:
    BeautifulSoup = None
    _BEAUTIFULSOUP_AVAILABLE = False


try:
    try:
        from ddgs import DDGS
    except ImportError:  # Fallback for older versions of ddgs
        from duckduckgo_search import DDGS
    _DDGS_AVAILABLE = True
except ImportError:
    DDGS = None
    _DDGS_AVAILABLE = False


__all__ = (
    'AbstractBuiltinTool',
    'WebSearchTool',
    'WebSearchUserLocation',
    'CodeExecutionTool',
    'UrlContextTool',
    'WebSearch',
    'WebRead'
)


@dataclass(kw_only=True)
class AbstractBuiltinTool(ABC):
    """Abstract base class for built-in tools provided by model providers."""
    
    kind: str = 'unknown_builtin_tool'
    """Built-in tool identifier, used as a discriminator."""


@dataclass(kw_only=True)
class WebSearchTool(AbstractBuiltinTool):
    """A built-in tool that allows models to search the web for information.
    
    The exact parameters supported depend on the model provider:
    
    Supported by:
    - Anthropic
    - OpenAI Responses
    - Groq
    - Google
    """
    
    search_context_size: Literal['low', 'medium', 'high'] = 'medium'
    """Controls how much context is retrieved from web searches.
    
    Supported by:
    - OpenAI Responses
    """
    
    user_location: Optional['WebSearchUserLocation'] = None
    """Localizes search results based on user location.
    
    Supported by:
    - Anthropic
    - OpenAI Responses
    """
    
    blocked_domains: Optional[list[str]] = None
    """Domains to exclude from search results.
    
    Note: With Anthropic, you can only use one of blocked_domains or allowed_domains.
    
    Supported by:
    - Anthropic
    - Groq
    """
    
    allowed_domains: Optional[list[str]] = None
    """If provided, only these domains will be included in results.
    
    Note: With Anthropic, you can only use one of blocked_domains or allowed_domains.
    
    Supported by:
    - Anthropic
    - Groq
    """
    
    max_uses: Optional[int] = None
    """Maximum number of web searches allowed.
    
    Supported by:
    - Anthropic
    """
    
    kind: str = 'web_search'
    """The kind of tool."""


class WebSearchUserLocation(TypedDict, total=False):
    """User location information for localizing web search results.
    
    Supported by:
    - Anthropic
    - OpenAI Responses
    """
    
    city: str
    """The city where the user is located."""
    
    country: str
    """The country where the user is located.
    For OpenAI, this must be a 2-letter country code (e.g., 'US', 'GB').
    """
    
    region: str
    """The region or state where the user is located."""
    
    timezone: str
    """The timezone of the user's location."""


@dataclass(kw_only=True)
class CodeExecutionTool(AbstractBuiltinTool):
    """A built-in tool that allows models to execute code.
    
    Supported by:
    - Anthropic
    - OpenAI Responses
    - Google
    """
    
    kind: str = 'code_execution'
    """The kind of tool."""


@dataclass(kw_only=True)
class UrlContextTool(AbstractBuiltinTool):
    """Allows models to access contents from URLs.
    
    Supported by:
    - Google
    """
    
    kind: str = 'url_context'
    """The kind of tool."""


def WebSearch(query: str, max_results: int = 10) -> str:
    """
    Search the web for the given query and return formatted results.

    Args:
        query: The search query
        max_results: Maximum number of results to return (default: 10)

    Returns:
        Formatted string containing search results
    """
    if not _DDGS_AVAILABLE:
        from upsonic.utils.printing import import_error
        import_error(
            package_name="duckduckgo-search",
            install_command='pip install "upsonic[tools]"',
            feature_name="DuckDuckGo search tool"
        )

    with DDGS() as ddgs:
        try:
            results = list(ddgs.text(query, max_results=max_results))
            
            formatted_results = f"Web search results for: {query}\n\n"
            for i, result in enumerate(results, 1):
                formatted_results += f"{i}. {result.get('title', 'No title')}\n"
                formatted_results += f"   URL: {result.get('href', 'No URL')}\n"
                formatted_results += f"   Description: {result.get('body', 'No description')}\n\n"
            
            return formatted_results
        except Exception as e:
            return f"Error performing web search: {str(e)}"


def WebRead(url: str) -> str:
    """
    Read and extract text content from a web page.

    Args:
        url: The URL to read from

    Returns:
        Extracted text content from the webpage
    """
    if not _REQUESTS_AVAILABLE:
        from upsonic.utils.printing import import_error
        import_error(
            package_name="requests",
            install_command='pip install "upsonic[loaders]"',
            feature_name="WebRead tool"
        )

    if not _BEAUTIFULSOUP_AVAILABLE:
        from upsonic.utils.printing import import_error
        import_error(
            package_name="beautifulsoup4",
            install_command='pip install "upsonic[loaders]"',
            feature_name="WebRead tool"
        )

    session = requests.Session()
    try:
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
        }
        
        response = session.get(url, headers=headers, timeout=30)
        response.raise_for_status()
        
        soup = BeautifulSoup(response.content, 'html.parser')
        
        for script in soup(["script", "style"]):
            script.decompose()
        
        text = soup.get_text()
        
        lines = (line.strip() for line in text.splitlines())
        chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
        text = ' '.join(chunk for chunk in chunks if chunk)
        
        if len(text) > 5000:
            text = text[:5000] + "... [Content truncated]"
        
        return f"Content from {url}:\n\n{text}"
    except requests.exceptions.RequestException as e:
        return f"Error reading from {url}: {str(e)}"
    except Exception as e:
        return f"Error processing content from {url}: {str(e)}"
    finally:
        session.close()


from __future__ import annotations

import json
import time
from typing import Optional, Dict, Any, Type, Union, TypeVar, TYPE_CHECKING

from pydantic import BaseModel

from upsonic.storage.base import Storage
from upsonic.storage.session.sessions import InteractionSession, UserProfile

if TYPE_CHECKING:
    from redis.asyncio import Redis
    from redis.exceptions import ConnectionError as RedisConnectionError

try:
    from redis.asyncio import Redis
    from redis.exceptions import ConnectionError as RedisConnectionError
    _REDIS_AVAILABLE = True
except ImportError:
    Redis = None  # type: ignore
    RedisConnectionError = None  # type: ignore
    _REDIS_AVAILABLE = False


T = TypeVar('T', bound=BaseModel)

class RedisStorage(Storage):
    """
    A hybrid sync/async, high-performance storage provider using Redis and
    its native async client, with proper connection lifecycle management.
    """

    def __init__(
        self,
        prefix: str,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: Optional[str] = None,
        ssl: bool = False,
        expire: Optional[int] = None,
    ):
        """
        Initializes the async Redis storage provider.

        Args:
            prefix: A prefix to namespace all keys for this application instance.
            host: The Redis server hostname.
            port: The Redis server port.
            db: The Redis database number to use.
            password: Optional password for Redis authentication.
            ssl: If True, uses an SSL connection.
            expire: Optional TTL in seconds for all created keys.
        """
        if not _REDIS_AVAILABLE:
            from upsonic.utils.printing import import_error
            import_error(
                package_name="redis",
                install_command='pip install "upsonic[storage]"',
                feature_name="Redis storage provider"
            )

        super().__init__()
        self.prefix = prefix
        self.expire = expire
        # Client is configured but not connected on initialization
        self.redis_client: Redis = Redis(
            host=host, port=port, db=db, password=password,
            ssl=ssl, decode_responses=True
        )

    def _get_key(self, object_id: str, model_type: Type[BaseModel]) -> str:
        if model_type is InteractionSession: return f"{self.prefix}:session:{object_id}"
        elif model_type is UserProfile: return f"{self.prefix}:profile:{object_id}"
        raise TypeError(f"Unsupported model type for key generation: {model_type.__name__}")
    
    def _serialize(self, data: Dict[str, Any]) -> str: return json.dumps(data)
    def _deserialize(self, data: str) -> Dict[str, Any]: return json.loads(data)



    def is_connected(self) -> bool: return self._run_async_from_sync(self.is_connected_async())
    def connect(self) -> None: return self._run_async_from_sync(self.connect_async())
    def disconnect(self) -> None: return self._run_async_from_sync(self.disconnect_async())
    def create(self) -> None: return self._run_async_from_sync(self.create_async())
    def read(self, object_id: str, model_type: Type[T]) -> Optional[T]: return self._run_async_from_sync(self.read_async(object_id, model_type))
    def upsert(self, data: Union[InteractionSession, UserProfile]) -> None: return self._run_async_from_sync(self.upsert_async(data))
    def delete(self, object_id: str, model_type: Type[BaseModel]) -> None: return self._run_async_from_sync(self.delete_async(object_id, model_type))
    def drop(self) -> None: return self._run_async_from_sync(self.drop_async())
    


    async def is_connected_async(self) -> bool:
        if not self._connected:
            return False
        try:
            await self.redis_client.ping()
            return True
        except (RedisConnectionError, ConnectionRefusedError):
            self._connected = False
            return False

    async def connect_async(self) -> None:
        if self._connected and await self.is_connected_async():
            return
        try:
            await self.redis_client.ping()
            self._connected = True
        except (RedisConnectionError, ConnectionRefusedError) as e:
            self._connected = False
            raise ConnectionError(f"Failed to connect to Redis: {e}") from e

    async def disconnect_async(self) -> None:
        await self.redis_client.close()
        self._connected = False

    async def create_async(self) -> None:
        await self.connect_async()

    async def read_async(self, object_id: str, model_type: Type[T]) -> Optional[T]:
        key = self._get_key(object_id, model_type)
        data_str = await self.redis_client.get(key)
        if data_str is None:
            return None
        try:
            data_dict = self._deserialize(data_str)
            return model_type.from_dict(data_dict)
        except (json.JSONDecodeError, TypeError) as e:
            from upsonic.utils.printing import warning_log
            warning_log(f"Could not parse key {key}. Error: {e}", "RedisStorage")
            return None

    async def upsert_async(self, data: Union[InteractionSession, UserProfile]) -> None:
        data.updated_at = time.time()
        data_dict = data.model_dump(mode="json")
        json_string = self._serialize(data_dict)

        if isinstance(data, InteractionSession): key = self._get_key(data.session_id, InteractionSession)
        elif isinstance(data, UserProfile): key = self._get_key(data.user_id, UserProfile)
        else: raise TypeError(f"Unsupported data type for upsert: {type(data).__name__}")
        
        await self.redis_client.set(key, json_string, ex=self.expire)

    async def delete_async(self, object_id: str, model_type: Type[BaseModel]) -> None:
        key = self._get_key(object_id, model_type)
        await self.redis_client.delete(key)

    async def drop_async(self) -> None:
        """Asynchronously deletes ALL keys associated with this provider's prefix."""
        keys_to_delete = [key async for key in self.redis_client.scan_iter(match=f"{self.prefix}:*")]
        if keys_to_delete:
            await self.redis_client.delete(*keys_to_delete)



from pydantic import BaseModel, Field
from typing import List, Optional, Dict

class UserTraits(BaseModel):
    """
    A comprehensive schema to capture a user's profile, preferences, and
    in-session state. This model is used by the AI to tailor its responses
    and maintain long-term context about the user.
    """
    
    detected_expertise: Optional[str] = Field(
        None, 
        description="The user's expertise level on the main topic of conversation, e.g., 'beginner', 'intermediate', 'expert'."
    )
    detected_tone: Optional[str] = Field(
        None,
        description="The user's preferred communication tone, inferred from their language, e.g., 'formal', 'casual', 'technical'."
    )
    inferred_interests: Optional[List[str]] = Field(
        None,
        description="A list of general topics or keywords the user seems interested in based on the dialogue."
    )

    current_goal: Optional[str] = Field(
        None,
        description="A concise summary of the user's immediate objective in this specific conversation. What are they trying to accomplish right now? E.g., 'debug a function', 'draft a marketing email', 'understand quantum mechanics'."
    )
    
    session_sentiment: Optional[str] = Field(
        None,
        description="The user's dominant emotional state in the last few messages. E.g., 'curious', 'frustrated', 'pleased', 'confused', 'neutral'."
    )

    communication_style: Optional[List[str]] = Field(
        None,
        description="Inferred preferences for how the user wants information presented. E.g., 'prefers concise answers', 'requests code examples', 'likes bullet points', 'enjoys detailed explanations', 'asks follow-up questions'."
    )
    
    key_entities: Optional[Dict[str, str]] = Field(
        None,
        description="A dictionary of important, specific named entities (people, projects, tools) mentioned by the user, along with their context. E.g., {'ProjectTitan': 'A key project the user is working on', 'JaneDoe': 'A colleague the user mentioned'}."
    )
    
    long_term_objective_summary: Optional[str] = Field(
        None,
        description="A running summary of the user's overarching goals or interests observed across multiple sessions. This should be updated, not replaced, to build a long-term understanding."
    )


from __future__ import annotations
import asyncio
import os
from typing import List, Dict, Any, Optional
import numpy as np


try:
    from transformers import AutoTokenizer, AutoModel, BitsAndBytesConfig
    import torch.nn.functional as F
    import torch
    from huggingface_hub import hf_hub_download, login, InferenceClient
    _TRANSFORMERS_AVAILABLE = True
except ImportError:
    AutoTokenizer = None
    AutoModel = None
    BitsAndBytesConfig = None
    F = None
    torch = None
    hf_hub_download = None
    login = None
    InferenceClient = None
    _TRANSFORMERS_AVAILABLE = False
from pydantic import Field, field_validator

from .base import EmbeddingProvider, EmbeddingConfig, EmbeddingMode
from upsonic.utils.printing import info_log, debug_log
from ..utils.package.exception import ConfigurationError, ModelConnectionError


class HuggingFaceEmbeddingConfig(EmbeddingConfig):
    """Configuration for HuggingFace embedding models."""
    
    model_name: str = Field("sentence-transformers/all-MiniLM-L6-v2", description="HuggingFace model name or path")
    
    hf_token: Optional[str] = Field(None, description="HuggingFace API token")
    
    use_api: bool = Field(False, description="Use HuggingFace Inference API instead of local model")
    use_local: bool = Field(True, description="Use local model execution")
    
    device: Optional[str] = Field(None, description="Device to run model on (auto-detected if None)")
    torch_dtype: str = Field("float32", description="PyTorch data type (float16, float32, bfloat16)")
    trust_remote_code: bool = Field(False, description="Trust remote code in model")
    
    max_seq_length: Optional[int] = Field(None, description="Maximum sequence length")
    pooling_strategy: str = Field("mean", description="Pooling strategy (mean, cls, max)")
    normalize_embeddings: bool = Field(True, description="Normalize embeddings to unit length")
    
    enable_quantization: bool = Field(False, description="Enable model quantization")
    quantization_bits: int = Field(8, description="Quantization bits (4, 8, 16)")
    enable_gradient_checkpointing: bool = Field(False, description="Enable gradient checkpointing to save memory")
    
    wait_for_model: bool = Field(True, description="Wait for model to load if using API")
    timeout: int = Field(None, description="Timeout for model")

    cache_dir: Optional[str] = Field(None, description="Model cache directory")
    force_download: bool = Field(False, description="Force re-download of model")
    
    @field_validator('pooling_strategy')
    @classmethod
    def validate_pooling_strategy(cls, v):
        valid_strategies = ['mean', 'cls', 'max', 'mean_sqrt_len']
        if v not in valid_strategies:
            raise ValueError(f"Invalid pooling strategy: {v}. Valid options: {valid_strategies}")
        return v
    
    @field_validator('torch_dtype')
    @classmethod
    def validate_torch_dtype(cls, v):
        valid_dtypes = ['float16', 'float32', 'bfloat16']
        if v not in valid_dtypes:
            raise ValueError(f"Invalid torch_dtype: {v}. Valid options: {valid_dtypes}")
        return v


class HuggingFaceEmbedding(EmbeddingProvider):

    
    config: HuggingFaceEmbeddingConfig
    
    def __init__(self, config: Optional[HuggingFaceEmbeddingConfig] = None, **kwargs):
        if not _TRANSFORMERS_AVAILABLE:
            from upsonic.utils.printing import import_error
            import_error(
                package_name="transformers",
                install_command='pip install transformers',
                feature_name="transformers provider"
            )

        if config is None:
            config = HuggingFaceEmbeddingConfig(**kwargs)

        super().__init__(config=config)
        
        self._setup_device()
        self._setup_authentication()
        
        if self.config.use_local and not self.config.use_api:
            self._setup_local_model()
        
        self._model_info: Optional[Dict[str, Any]] = None
        
        if self.config.use_api:
            self._setup_api_session()
    
    def _setup_device(self):
        """Setup compute device for local models."""
        if self.config.device is None:
            if torch.cuda.is_available():
                self.device = "cuda"
            elif hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
                self.device = "mps"
            else:
                self.device = "cpu"
        else:
            self.device = self.config.device
            
        info_log(f"HuggingFace embedding using device: {self.device}", context="HuggingFaceEmbedding")
    
    def _setup_authentication(self):
        """Setup HuggingFace authentication."""
        hf_token = self.config.hf_token or os.getenv("HF_TOKEN") or os.getenv("HUGGINGFACE_HUB_TOKEN")
        
        if hf_token:
            try:
                login(token=hf_token)
                from upsonic.utils.printing import success_log
                success_log("Successfully authenticated with HuggingFace Hub", "HuggingFaceProvider")
            except Exception as e:
                from upsonic.utils.printing import warning_log
                warning_log(f"HuggingFace authentication failed: {e}", "HuggingFaceProvider")
    
    def _setup_local_model(self):
        """Setup local model and tokenizer with optional quantization."""
        
        try:
            info_log(f"Loading HuggingFace model: {self.config.model_name}", context="HuggingFaceEmbedding")
            
            model_kwargs = {
                "trust_remote_code": self.config.trust_remote_code,
                "cache_dir": self.config.cache_dir,
                "force_download": self.config.force_download
            }
            
            
            torch_dtype = getattr(torch, self.config.torch_dtype)
            
            if self.config.enable_quantization:
                if self.config.quantization_bits == 8:
                    debug_log("Configuring 8-bit quantization using bitsandbytes...", context="HuggingFaceEmbedding")
                    model_kwargs['load_in_8bit'] = True
                    model_kwargs['device_map'] = "auto"

                elif self.config.quantization_bits == 4:
                    debug_log("Configuring 4-bit quantization using bitsandbytes...", context="HuggingFaceEmbedding")
                    quantization_config = BitsAndBytesConfig(
                        load_in_4bit=True,
                        bnb_4bit_quant_type="nf4",
                        bnb_4bit_use_double_quant=True,
                        bnb_4bit_compute_dtype=torch.bfloat16
                    )
                    model_kwargs['quantization_config'] = quantization_config
                    model_kwargs['device_map'] = "auto"
                    torch_dtype = torch.bfloat16

            tokenizer_kwargs = {k: v for k, v in model_kwargs.items() if k not in ['quantization_config', 'device_map', 'load_in_8bit']}
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.config.model_name,
                **tokenizer_kwargs
            )
            
            self.model = AutoModel.from_pretrained(
                self.config.model_name,
                torch_dtype=torch_dtype,
                **model_kwargs
            )

            if 'device_map' not in model_kwargs:
                self.model = self.model.to(self.device)
                debug_log(f"Model moved to device: {self.device}", context="HuggingFaceEmbedding")

            self.model.eval()
            
            if self.config.enable_gradient_checkpointing:
                self.model.gradient_checkpointing_enable()
            
            info_log("Model loaded successfully.", context="HuggingFaceEmbedding")
            
        except Exception as e:
            if "bitsandbytes" in str(e).lower():
                raise ConfigurationError(
                    "Quantization requires the 'bitsandbytes' and 'accelerate' libraries. Please install them with 'pip install bitsandbytes accelerate'",
                    error_code="DEPENDENCY_MISSING",
                    original_error=e
                )
            raise ConfigurationError(
                f"Failed to load HuggingFace model {self.config.model_name}: {str(e)}",
                error_code="MODEL_LOAD_ERROR",
                original_error=e
            )
    
    def _setup_api_session(self):
        """Setup InferenceClient for HuggingFace API calls."""
        
        hf_token = self.config.hf_token or os.getenv("HF_TOKEN") or os.getenv("HUGGINGFACE_HUB_TOKEN")
        
        self._inference_client = InferenceClient(
            token=hf_token,
            timeout=getattr(self.config, 'timeout', 30)
        )

    
    @property
    def supported_modes(self) -> List[EmbeddingMode]:
        """HuggingFace supports all embedding modes."""
        return [EmbeddingMode.DOCUMENT, EmbeddingMode.QUERY, EmbeddingMode.SYMMETRIC]
    
    @property 
    def pricing_info(self) -> Dict[str, float]:
        """Get HuggingFace pricing info (API usage)."""
        if self.config.use_local:
            return {
                "per_million_tokens": 0.0,
                "currency": "USD",
                "note": "Local model execution - no API costs"
            }
        else:
            return {
                "per_million_tokens": 0.01,
                "currency": "USD",
                "note": "HuggingFace Inference API pricing varies by model",
                "updated": "2024-01-01"
            }
    
    def get_model_info(self) -> Dict[str, Any]:
        """Get information about the current HuggingFace model."""
        if self._model_info is None:
            info = {
                "model_name": self.config.model_name,
                "provider": "HuggingFace",
                "type": "embedding",
                "execution_mode": "API" if self.config.use_api else "Local",
                "device": self.device if hasattr(self, 'device') else "unknown"
            }
            
            if hasattr(self, 'model') and self.model:
                try:
                    info["parameters"] = sum(p.numel() for p in self.model.parameters())
                    info["hidden_size"] = getattr(self.model.config, 'hidden_size', 'unknown')
                    info["max_position_embeddings"] = getattr(self.model.config, 'max_position_embeddings', 'unknown')
                except:
                    pass
            
            if hasattr(self, 'tokenizer') and self.tokenizer:
                info["vocab_size"] = len(self.tokenizer.vocab) if hasattr(self.tokenizer, 'vocab') else 'unknown'
                info["model_max_length"] = getattr(self.tokenizer, 'model_max_length', 'unknown')
            
            self._model_info = info
        
        return self._model_info
    
    def _mean_pooling(self, model_output, attention_mask):
        """Apply mean pooling to get sentence embeddings."""
        token_embeddings = model_output[0]
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    
    def _cls_pooling(self, model_output, attention_mask):
        """Use CLS token for pooling."""
        return model_output[0][:, 0]
    
    def _max_pooling(self, model_output, attention_mask):
        """Apply max pooling."""
        token_embeddings = model_output[0]
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        token_embeddings[input_mask_expanded == 0] = -1e9
        return torch.max(token_embeddings, 1)[0]
    
    def _apply_pooling(self, model_output, attention_mask):
        """Apply the configured pooling strategy."""
        if self.config.pooling_strategy == "mean":
            return self._mean_pooling(model_output, attention_mask)
        elif self.config.pooling_strategy == "cls":
            return self._cls_pooling(model_output, attention_mask)
        elif self.config.pooling_strategy == "max":
            return self._max_pooling(model_output, attention_mask)
        else:
            return self._mean_pooling(model_output, attention_mask)
    
    async def _embed_local(self, texts: List[str]) -> List[List[float]]:
        """Embed texts using local model."""
        if not hasattr(self, 'model') or not hasattr(self, 'tokenizer'):
            raise ModelConnectionError("Local model not initialized")
        
        try:
            max_length = self.config.max_seq_length or self.tokenizer.model_max_length
            if max_length > 512:
                max_length = 512
            
            encoded_input = self.tokenizer(
                texts,
                padding=True,
                truncation=True,
                max_length=max_length,
                return_tensors='pt'
            )
            
            encoded_input = {key: value.to(self.device) for key, value in encoded_input.items()}
            
            with torch.no_grad():
                model_output = self.model(**encoded_input)
                
                sentence_embeddings = self._apply_pooling(model_output, encoded_input['attention_mask'])
                
                if self.config.normalize_embeddings:
                    sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)
                
                embeddings = sentence_embeddings.cpu().numpy().tolist()
                
                return embeddings
                
        except Exception as e:
            raise ModelConnectionError(
                f"Local embedding failed: {str(e)}",
                error_code="LOCAL_EMBEDDING_ERROR",
                original_error=e
            )
    

    async def _embed_api(self, texts: List[str]) -> List[List[float]]:
        """Embed texts using HuggingFace InferenceClient."""
        if not self._inference_client:
            raise ModelConnectionError("InferenceClient not initialized")
        
        try:
            def sync_api_call():
                return self._inference_client.feature_extraction(
                    text=texts,
                    model=self.config.model_name
                )

            embeddings_np = await asyncio.to_thread(sync_api_call)

            if isinstance(embeddings_np, np.ndarray):
                if embeddings_np.ndim == 1:
                    return [embeddings_np.tolist()]
                else:
                    return embeddings_np.tolist()
            
            if isinstance(embeddings_np, list):
                return embeddings_np
            
            return []
            
        except Exception as e:
            raise ModelConnectionError(f"Failed to get embeddings via API: {str(e)}", original_error=e)
    
    async def _embed_batch(self, texts: List[str], mode: EmbeddingMode = EmbeddingMode.DOCUMENT) -> List[List[float]]:
        """
        Embed a batch of texts using HuggingFace model or API.
        
        Args:
            texts: List of text strings to embed
            mode: Embedding mode
            
        Returns:
            List of embedding vectors
        """
        if not texts:
            return []
        
        if self.config.use_api:
            return await self._embed_api(texts)
        else:
            return await self._embed_local(texts)
    
    async def validate_connection(self) -> bool:
        """Validate HuggingFace model or API connection."""
        try:
            test_result = await self._embed_batch(["test connection"], EmbeddingMode.QUERY)
            return len(test_result) > 0 and len(test_result[0]) > 0
            
        except Exception as e:
            debug_log(f"HuggingFace connection validation failed: {str(e)}", context="HuggingFaceEmbedding")
            return False
    
    def get_memory_usage(self) -> Dict[str, Any]:
        """Get memory usage information for local models."""
        if not hasattr(self, 'model') or self.device == "cpu":
            return {"memory_usage": "N/A", "device": self.device}
        
        try:
            if self.device.startswith("cuda"):
                allocated = torch.cuda.memory_allocated(self.device)
                cached = torch.cuda.memory_reserved(self.device)
                
                return {
                    "device": self.device,
                    "allocated_memory_mb": allocated / 1024 / 1024,
                    "cached_memory_mb": cached / 1024 / 1024,
                    "total_memory_mb": torch.cuda.get_device_properties(self.device).total_memory / 1024 / 1024
                }
            else:
                return {"memory_usage": "Memory tracking not available for this device", "device": self.device}
                
        except Exception as e:
            return {"memory_usage": f"Error getting memory info: {e}", "device": self.device}

    async def close(self):
        """
        Clean up HuggingFace models, tokenizer, and API client.
        """
        if hasattr(self, 'model') and self.model:
            try:
                del self.model
                self.model = None
            except Exception as e:
                from upsonic.utils.printing import warning_log
                warning_log(f"Error closing HuggingFace model: {e}", "HuggingFaceProvider")
        
        if hasattr(self, 'tokenizer') and self.tokenizer:
            try:
                del self.tokenizer
                self.tokenizer = None
            except Exception as e:
                from upsonic.utils.printing import warning_log
                warning_log(f"Error closing HuggingFace tokenizer: {e}", "HuggingFaceProvider")
        
        if hasattr(self, '_inference_client') and self._inference_client:
            try:
                if hasattr(self._inference_client, 'aclose'):
                    await self._inference_client.aclose()
                elif hasattr(self._inference_client, 'close'):
                    if asyncio.iscoroutinefunction(self._inference_client.close):
                        await self._inference_client.close()
                    else:
                        self._inference_client.close()
                
                del self._inference_client
                self._inference_client = None
            except Exception as e:
                from upsonic.utils.printing import warning_log
                warning_log(f"Error closing HuggingFace InferenceClient: {e}", "HuggingFaceProvider")
        
        # Clear CUDA cache if using GPU
        if hasattr(self, 'device') and self.device.startswith('cuda'):
            try:
                import torch
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
                    from upsonic.utils.printing import success_log
                    success_log("CUDA cache cleared", "HuggingFaceProvider")
            except Exception as e:
                from upsonic.utils.printing import warning_log
                warning_log(f"Could not clear CUDA cache: {e}", "HuggingFaceProvider")
        
        await super().close()

    async def remove_local_cache(self) -> bool:
        """
        Remove HuggingFace model/tokenizer cache files from local storage.
        """
        from huggingface_hub import try_to_load_from_cache
        import shutil, os

        try:
            model_path = try_to_load_from_cache(
                repo_id=self.config.model_name,
                cache_dir=self.config.cache_dir
            )
            if model_path and os.path.exists(model_path):
                root = os.path.dirname(model_path)
                shutil.rmtree(root, ignore_errors=True)
                from upsonic.utils.printing import success_log
                success_log(f"Local cache removed for {self.config.model_name} at {root}", "HuggingFaceProvider")
                return True
            else:
                from upsonic.utils.printing import info_log
                info_log(f"No local cache found for {self.config.model_name}", "HuggingFaceProvider")
                return False
        except Exception as e:
            from upsonic.utils.printing import error_log
            error_log(f"Error removing local cache for {self.config.model_name}: {e}", "HuggingFaceProvider")
            return False


def create_sentence_transformer_embedding(
    model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
    **kwargs
) -> HuggingFaceEmbedding:
    """Create a sentence transformer embedding provider."""
    config = HuggingFaceEmbeddingConfig(
        model_name=model_name,
        use_local=True,
        pooling_strategy="mean",
        normalize_embeddings=True,
        **kwargs
    )
    return HuggingFaceEmbedding(config=config)


def create_mpnet_embedding(**kwargs) -> HuggingFaceEmbedding:
    """Create MPNet embedding provider (high quality)."""
    return create_sentence_transformer_embedding(
        model_name="sentence-transformers/all-mpnet-base-v2",
        **kwargs
    )


def create_minilm_embedding(**kwargs) -> HuggingFaceEmbedding:
    """Create MiniLM embedding provider (fast and efficient)."""
    return create_sentence_transformer_embedding(
        model_name="sentence-transformers/all-MiniLM-L6-v2",
        **kwargs
    )


def create_huggingface_api_embedding(
    model_name: str,
    hf_token: Optional[str] = None,
    **kwargs
) -> HuggingFaceEmbedding:
    """Create HuggingFace API embedding provider."""
    config = HuggingFaceEmbeddingConfig(
        model_name=model_name,
        use_api=True,
        use_local=False,
        hf_token=hf_token,
        **kwargs
    )
    return HuggingFaceEmbedding(config=config)



import asyncio
import pytest
import time
from contextlib import asynccontextmanager
from unittest.mock import AsyncMock, MagicMock, patch, Mock
from typing import List, Dict, Any, Optional

from upsonic.agent.agent import Agent
from upsonic.agent.run_result import StreamRunResult
from upsonic.tasks.tasks import Task
from upsonic.storage.memory.memory import Memory
from upsonic.storage.providers.in_memory import InMemoryStorage
from upsonic.models import Model
from upsonic.messages.messages import (
    ModelRequest, ModelResponse, TextPart, PartStartEvent, 
    PartDeltaEvent, FinalResultEvent, UserPromptPart, SystemPromptPart
)


class MockModel(Model):
    """Mock model for testing streaming functionality."""
    
    def __init__(self, model_name: str = "test-model"):
        super().__init__()
        self._model_name = model_name
    
    @property
    def model_name(self) -> str:
        return self._model_name
    
    @property
    def system(self) -> str:
        return "test-provider"
    
    async def request(
        self,
        messages: list,
        model_settings: Any,
        model_request_parameters: Any,
    ) -> ModelResponse:
        """Mock request method."""
        return ModelResponse(
            parts=[TextPart(content="Hello world!")],
            model_name=self._model_name,
            timestamp=time.time(),
            usage=Mock(),
            provider_name="test-provider",
            provider_response_id="test-id",
            provider_details={},
            finish_reason="stop"
        )
    
    @asynccontextmanager
    async def request_stream(
        self,
        messages: list,
        model_settings: Any,
        model_request_parameters: Any,
    ):
        """Mock streaming request that yields test events."""
        # Create mock streaming events
        events = [
            PartStartEvent(index=0, part=TextPart(content="Hello")),
            PartDeltaEvent(index=0, delta=Mock(content_delta=" world")),
            PartDeltaEvent(index=0, delta=Mock(content_delta="!")),
            FinalResultEvent(tool_name=None, tool_call_id=None)
        ]
        
        # Create a mock stream context manager
        stream_mock = AsyncMock()
        stream_mock.__aenter__ = AsyncMock(return_value=stream_mock)
        stream_mock.__aexit__ = AsyncMock(return_value=None)
        
        async def mock_stream(self):
            for event in events:
                yield event
        
        stream_mock.__aiter__ = mock_stream
        stream_mock.get = Mock(return_value=ModelResponse(
            parts=[TextPart(content="Hello world!")],
            model_name=self._model_name,
            timestamp=time.time(),
            usage=Mock(),
            provider_name="test-provider",
            provider_response_id="test-id",
            provider_details={},
            finish_reason="stop"
        ))
        
        try:
            yield stream_mock
        finally:
            pass


class TestAgentStreaming:
    """Test suite for Agent streaming functionalities."""
    
    @pytest.fixture
    def mock_model(self):
        """Create a mock model for testing."""
        return MockModel()
    
    @pytest.fixture
    def agent(self, mock_model):
        """Create an agent instance for testing."""
        return Agent(model=mock_model, name="TestAgent")
    
    @pytest.fixture
    def simple_task(self):
        """Create a simple task for testing."""
        return Task(description="Hello, world!")
    
    @pytest.fixture
    def memory_storage(self):
        """Create in-memory storage for testing."""
        return InMemoryStorage()
    
    @pytest.fixture
    def memory(self, memory_storage):
        """Create memory instance for testing."""
        return Memory(
            storage=memory_storage,
            session_id="test-session",
            full_session_memory=False,  # Disable to avoid serialization issues
            summary_memory=False
        )
    
    def test_stream_returns_stream_run_result(self, agent, simple_task):
        """Test that stream() returns a StreamRunResult instance."""
        result = agent.stream(simple_task)
        
        assert isinstance(result, StreamRunResult)
        assert result._agent == agent
        assert result._task == simple_task
        assert result._debug is False
        assert result._retry == 1
    
    def test_stream_with_custom_parameters(self, agent, simple_task):
        """Test stream() with custom parameters."""
        result = agent.stream(
            simple_task, 
            model="custom-model", 
            debug=True, 
            retry=3
        )
        
        assert result._model == "custom-model"
        assert result._debug is True
        assert result._retry == 3
    
    def test_stream_initializes_task_properties(self, agent, simple_task):
        """Test that stream() properly initializes task properties."""
        # Ensure task properties are reset
        original_price_id = "existing-id"
        simple_task.price_id_ = original_price_id
        simple_task._tool_calls = [{"test": "call"}]

        result = agent.stream(simple_task)

        # Task should have a new price_id_ (not the original) and tool calls should be reset
        assert simple_task.price_id_ != original_price_id
        assert simple_task.price_id_ is not None  # Should have a new UUID
        assert simple_task._tool_calls == []
    
    @pytest.mark.asyncio
    async def test_stream_async_returns_stream_run_result(self, agent, simple_task):
        """Test that stream_async() returns a StreamRunResult instance."""
        result = await agent.stream_async(simple_task)
        
        assert isinstance(result, StreamRunResult)
        assert result._agent == agent
        assert result._task == simple_task
    
    @pytest.mark.asyncio
    async def test_stream_async_with_state_and_graph_id(self, agent, simple_task):
        """Test stream_async() with state and graph_execution_id."""
        mock_state = {"test": "state"}
        graph_id = "test-graph-id"
        
        result = await agent.stream_async(
            simple_task, 
            state=mock_state, 
            graph_execution_id=graph_id
        )
        
        assert result._state == mock_state
        assert result._graph_execution_id == graph_id
    
    @pytest.mark.asyncio
    async def test_stream_run_result_context_manager(self, agent, simple_task):
        """Test StreamRunResult as async context manager."""
        result = await agent.stream_async(simple_task)
        
        # Test context manager entry
        async with result as stream_result:
            assert stream_result._context_entered is True
            assert stream_result._start_time is not None
            assert stream_result is result
        
        # Test context manager exit
        assert result._context_entered is False
        assert result._end_time is not None
    
    @pytest.mark.asyncio
    async def test_stream_run_result_double_context_error(self, agent, simple_task):
        """Test that entering context manager twice raises error."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            with pytest.raises(RuntimeError, match="context manager is already active"):
                async with result:
                    pass
    
    @pytest.mark.asyncio
    async def test_stream_output_basic_functionality(self, agent, simple_task):
        """Test basic streaming output functionality."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            text_chunks = []
            async for chunk in result.stream_output():
                text_chunks.append(chunk)
            
            # Should have received text chunks
            assert len(text_chunks) > 0
            assert "".join(text_chunks) == "Hello world!"
    
    @pytest.mark.asyncio
    async def test_stream_output_outside_context_error(self, agent, simple_task):
        """Test that stream_output() outside context raises error."""
        result = await agent.stream_async(simple_task)
        
        with pytest.raises(RuntimeError, match="must be called within async context manager"):
            async for chunk in result.stream_output():
                pass
    
    @pytest.mark.asyncio
    async def test_stream_output_without_agent_task_error(self, agent, simple_task):
        """Test stream_output() without agent/task raises error."""
        result = StreamRunResult()
        result._context_entered = True
        
        with pytest.raises(RuntimeError, match="No agent or task available"):
            async for chunk in result.stream_output():
                pass
    
    @pytest.mark.asyncio
    async def test_stream_events_basic_functionality(self, agent, simple_task):
        """Test basic streaming events functionality."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            events = []
            async for event in result.stream_events():
                events.append(event)
            
            # Should have received various event types
            assert len(events) > 0
            event_types = [type(event).__name__ for event in events]
            assert "PartStartEvent" in event_types
            assert "FinalResultEvent" in event_types
    
    @pytest.mark.asyncio
    async def test_stream_events_outside_context_error(self, agent, simple_task):
        """Test that stream_events() outside context raises error."""
        result = await agent.stream_async(simple_task)
        
        with pytest.raises(RuntimeError, match="must be called within async context manager"):
            async for event in result.stream_events():
                pass
    
    @pytest.mark.asyncio
    async def test_get_final_output(self, agent, simple_task):
        """Test getting final output from stream result."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        final_output = result.get_final_output()
        assert final_output == "Hello world!"
    
    @pytest.mark.asyncio
    async def test_accumulated_text_property(self, agent, simple_task):
        """Test the accumulated text property."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        assert result.output == "Hello world!"
        assert result.get_accumulated_text() == "Hello world!"
    
    @pytest.mark.asyncio
    async def test_is_complete_status(self, agent, simple_task):
        """Test completion status tracking."""
        result = await agent.stream_async(simple_task)
        
        assert result.is_complete() is False
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        assert result.is_complete() is True
    
    @pytest.mark.asyncio
    async def test_streaming_events_tracking(self, agent, simple_task):
        """Test that streaming events are properly tracked."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for event in result.stream_events():
                pass  # Consume all events
        
        events = result.get_streaming_events()
        assert len(events) > 0
        
        # Test event filtering
        text_events = result.get_text_events()
        tool_events = result.get_tool_events()
        
        assert len(text_events) > 0
        assert len(tool_events) == 0  # No tool events in our mock
    
    @pytest.mark.asyncio
    async def test_streaming_stats(self, agent, simple_task):
        """Test streaming statistics collection."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        stats = result.get_streaming_stats()
        
        assert "total_events" in stats
        assert "text_events" in stats
        assert "tool_events" in stats
        assert "accumulated_chars" in stats
        assert "is_complete" in stats
        assert "has_final_output" in stats
        assert "event_types" in stats
        
        assert stats["is_complete"] is True
        assert stats["has_final_output"] is True
        assert stats["accumulated_chars"] == len("Hello world!")
    
    @pytest.mark.asyncio
    async def test_performance_metrics(self, agent, simple_task):
        """Test performance metrics collection."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        metrics = result.get_performance_metrics()
        
        assert "start_time" in metrics
        assert "end_time" in metrics
        assert "first_token_time" in metrics
        assert "total_duration" in metrics
        assert "time_to_first_token" in metrics
        assert "tokens_per_second" in metrics
        assert "characters_per_second" in metrics
        
        # All timing metrics should be present
        assert metrics["start_time"] is not None
        assert metrics["end_time"] is not None
        assert metrics["total_duration"] is not None
        assert metrics["total_duration"] > 0
    
    @pytest.mark.asyncio
    async def test_message_tracking(self, agent, simple_task):
        """Test message tracking in streaming."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        # Test message methods
        all_messages = result.all_messages()
        new_messages = result.new_messages()
        
        # Should have messages after streaming
        assert len(all_messages) > 0
        assert len(new_messages) > 0
    
    @pytest.mark.asyncio
    async def test_print_stream_async(self, agent, simple_task, capsys):
        """Test print_stream_async functionality."""
        result = await agent.print_stream_async(simple_task)
        
        # Should return the final output
        assert result == "Hello world!"
        
        # Should have printed the output (captured by capsys)
        captured = capsys.readouterr()
        assert "Hello world!" in captured.out
    
    def test_print_stream_synchronous(self, agent, simple_task, capsys):
        """Test print_stream synchronous functionality."""
        result = agent.print_stream(simple_task)
        
        # Should return the final output
        assert result == "Hello world!"
        
        # Should have printed the output
        captured = capsys.readouterr()
        assert "Hello world!" in captured.out
    
    @pytest.mark.asyncio
    async def test_streaming_with_memory(self, agent, simple_task, memory):
        """Test streaming with memory integration."""
        agent.memory = memory
        
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        # Memory should have been updated
        assert result.get_final_output() == "Hello world!"
    
    @pytest.mark.asyncio
    async def test_streaming_with_tools(self, agent, simple_task):
        """Test streaming with tools enabled."""
        # Add a simple tool to the task
        def test_tool() -> str:
            """A simple test tool that returns a string."""
            return "Tool executed"

        simple_task.tools = [test_tool]
        
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        # Should complete successfully even with tools
        assert result.get_final_output() is not None
    
    @pytest.mark.asyncio
    async def test_streaming_error_handling(self, agent, simple_task):
        """Test error handling in streaming."""
        # Create a model that raises an error
        error_model = Mock()
        
        # Create a proper async context manager that raises an error
        class ErrorContextManager:
            async def __aenter__(self):
                raise Exception("Stream error")
            
            async def __aexit__(self, exc_type, exc_val, exc_tb):
                pass
        
        error_model.request_stream = lambda *args, **kwargs: ErrorContextManager()
        error_model.model_name = "error-model"
        error_model.settings = None
        error_model.customize_request_parameters = Mock(return_value={})

        # Mock the infer_model function to return our error model
        with patch('upsonic.models.infer_model', return_value=error_model):
            agent.model = error_model
            result = await agent.stream_async(simple_task)

            with pytest.raises(Exception, match="Stream error"):
                async with result:
                    async for chunk in result.stream_output():
                        pass
    
    @pytest.mark.asyncio
    async def test_streaming_with_caching(self, agent, simple_task):
        """Test streaming with caching enabled."""
        simple_task.enable_cache = True
        
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        # Should complete successfully with caching
        assert result.get_final_output() == "Hello world!"
    
    @pytest.mark.asyncio
    async def test_streaming_with_guardrails(self, agent, simple_task):
        """Test streaming with guardrails."""
        def guardrail(text):
            return len(text) > 0, text
        
        simple_task.guardrail = guardrail
        
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        # Should complete successfully with guardrails
        assert result.get_final_output() == "Hello world!"
    
    @pytest.mark.asyncio
    async def test_streaming_with_compression(self, agent, simple_task):
        """Test streaming with context compression."""
        agent.compression_strategy = "simple"
        agent.compression_settings = {"max_length": 100}
        
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        # Should complete successfully with compression
        assert result.get_final_output() == "Hello world!"
    
    def test_stream_run_result_string_representation(self, agent, simple_task):
        """Test string representation of StreamRunResult."""
        result = agent.stream(simple_task)
        
        # Test __str__ method
        str_repr = str(result)
        assert str_repr == ""  # Empty before streaming
        
        # Test __repr__ method
        repr_str = repr(result)
        assert "StreamRunResult" in repr_str
        assert "streaming" in repr_str or "complete" in repr_str
    
    @pytest.mark.asyncio
    async def test_stream_run_result_repr_after_completion(self, agent, simple_task):
        """Test __repr__ after streaming completion."""
        result = await agent.stream_async(simple_task)
        
        async with result:
            async for chunk in result.stream_output():
                pass  # Consume all chunks
        
        repr_str = repr(result)
        assert "StreamRunResult" in repr_str
        assert "complete" in repr_str
        assert "accumulated_chars" in repr_str
        assert "events_count" in repr_str
        assert "messages_count" in repr_str


class TestAgentStreamingIntegration:
    """Integration tests for Agent streaming with various components."""
    
    @pytest.fixture
    def mock_model(self):
        """Create a mock model for integration testing."""
        return MockModel()
    
    @pytest.fixture
    def agent_with_memory(self, mock_model):
        """Create an agent with memory for integration testing."""
        storage = InMemoryStorage()
        memory = Memory(
            storage=storage,
            session_id="integration-test",
            full_session_memory=False,  # Disable to avoid serialization issues
            summary_memory=False
        )
        
        return Agent(
            model=mock_model,
            name="IntegrationAgent",
            memory=memory
        )
    
    @pytest.fixture
    def complex_task(self):
        """Create a complex task for integration testing."""
        def process_tool(x: str) -> str:
            """Process input and return result."""
            return f"Processed: {x}"
        
        return Task(
            description="Analyze this data and provide insights",
            tools=[process_tool],
            enable_cache=False,  # Disable cache to avoid embedding provider issues
            response_format=str
        )
    
    @pytest.mark.asyncio
    async def test_full_streaming_workflow(self, agent_with_memory, complex_task):
        """Test complete streaming workflow with all features."""
        result = await agent_with_memory.stream_async(complex_task)
        
        async with result:
            # Test both text and event streaming
            text_chunks = []
            events = []
            
            # Stream text
            async for chunk in result.stream_output():
                text_chunks.append(chunk)
            
            # Stream events
            async for event in result.stream_events():
                events.append(event)
        
        # Verify results
        assert result.get_final_output() == "Hello world!"
        assert len(text_chunks) > 0
        assert len(events) > 0
        assert result.is_complete() is True
        
        # Verify performance metrics
        metrics = result.get_performance_metrics()
        assert metrics["total_duration"] > 0
        
        # Verify streaming stats
        stats = result.get_streaming_stats()
        assert stats["is_complete"] is True
        assert stats["has_final_output"] is True
    
    @pytest.mark.asyncio
    async def test_multiple_streaming_sessions(self, agent_with_memory):
        """Test multiple streaming sessions with the same agent."""
        task1 = Task(description="First task")
        task2 = Task(description="Second task")
        
        # First streaming session
        result1 = await agent_with_memory.stream_async(task1)
        async with result1:
            async for chunk in result1.stream_output():
                pass
        
        # Second streaming session
        result2 = await agent_with_memory.stream_async(task2)
        async with result2:
            async for chunk in result2.stream_output():
                pass
        
        # Both should complete successfully
        assert result1.get_final_output() == "Hello world!"
        assert result2.get_final_output() == "Hello world!"
        assert result1.is_complete() is True
        assert result2.is_complete() is True
    
    @pytest.mark.asyncio
    async def test_streaming_with_multiple_tasks(self, agent_with_memory):
        """Test multiple streaming tasks sequentially."""
        tasks = [
            Task(description=f"Task {i}") 
            for i in range(3)
        ]
        
        results = []
        for task in tasks:
            result = await agent_with_memory.stream_async(task)
            async with result:
                async for chunk in result.stream_output():
                    pass  # Consume all chunks
            results.append(result.get_final_output())
        
        # All should complete successfully
        assert all(result == "Hello world!" for result in results)
        assert len(results) == 3


if __name__ == "__main__":
    pytest.main([__file__, "-v"])
